import plotly.io as pio

import peyes
from analysis._default_values._helpers import *
import analysis.statistics.matched_sdt as msdt

pio.renderers.default = "browser"

THRESHOLD = 5  # samples
MATCHING_SCHEME = "window"
METRICS = [peyes.constants.MATCH_RATIO_STR, peyes.constants.F1_STR, peyes.constants.D_PRIME_STR]

# %%
######################
## Fixation Matches ##

LABEL = 1       # EventLabelEnum.FIXATION.value

matched_fixations_sdt = msdt.load(
    dataset_name=DATASET_NAME, output_dir=PROCESSED_DATA_DIR,
    label=LABEL, stimulus_type=STIMULUS_TYPE, matching_schemes=None, metrics=None
)

fixation_statistics, fixation_pvalues, fixation_dunns, fixation_Ns = msdt.kruskal_wallis_dunns(
    matched_fixations_sdt,
    f"{MATCHING_SCHEME}_{THRESHOLD}",
    [GT1, GT2],
    metrics=METRICS,
    multi_comp=MULTI_COMP
)

### Show Figures

fixation_single_threshold_fig = msdt.single_scheme_figure(
    matched_fixations_sdt, f"{MATCHING_SCHEME}_{THRESHOLD}", GT1, gt2=GT2, metrics=METRICS, only_box=False
)
fixation_single_threshold_fig.show()

fixation_multi_threshold_figs = msdt.multi_metric_figure(
    matched_fixations_sdt,
    MATCHING_SCHEME,
    [peyes.constants.MATCH_RATIO_STR, peyes.constants.F1_STR, peyes.constants.D_PRIME_STR],
    show_other_gt=True, show_err_bands=True
)
fixation_multi_threshold_figs.show()

# %%
#####################
## Saccade Matches ##

LABEL = 2       # EventLabelEnum.SACCADE.value

matched_saccade_sdt = msdt.load(
    dataset_name=DATASET_NAME, output_dir=PROCESSED_DATA_DIR,
    label=LABEL, stimulus_type=STIMULUS_TYPE, matching_schemes=None, metrics=None
)

saccade_statistics, saccade_pvalues, saccade_dunns, saccade_Ns = msdt.kruskal_wallis_dunns(
    matched_fixations_sdt,
    f"{MATCHING_SCHEME}_{THRESHOLD}",
    [GT1, GT2],
    metrics=METRICS,
    multi_comp=MULTI_COMP
)

### Show Figures

saccade_single_threshold_fig = msdt.single_scheme_figure(
    matched_fixations_sdt, f"{MATCHING_SCHEME}_{THRESHOLD}", GT1, gt2=GT2, metrics=METRICS, only_box=False
)
saccade_single_threshold_fig.show()

saccade_multi_threshold_figs = msdt.multi_metric_figure(
    matched_fixations_sdt,
    MATCHING_SCHEME,
    [peyes.constants.MATCH_RATIO_STR, peyes.constants.F1_STR, peyes.constants.D_PRIME_STR],
    show_other_gt=True, show_err_bands=True
)
saccade_multi_threshold_figs.show()

# %%
######################
## Save Figures

peyes.visualize.save_figure(
    fixation_single_threshold_fig, "single_threshold-fixation", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)
peyes.visualize.save_figure(
    fixation_multi_threshold_figs, "multi_threshold-fixation", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)

peyes.visualize.save_figure(
    saccade_single_threshold_fig, "single_threshold-saccade", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)
peyes.visualize.save_figure(
    saccade_multi_threshold_figs, "multi_threshold-saccade", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)




## NUMERICAL CONSTANTS ##
####  Time  ####
SECONDS_PER_MINUTE = MINUTES_PER_HOUR = 60
MICROSECONDS_PER_MILLISECOND = MILLISECONDS_PER_SECOND = 1000
MICROSECONDS_PER_SECOND = MICROSECONDS_PER_MILLISECOND * MILLISECONDS_PER_SECOND  # 1,000,000

####  Tobii Monitor  ####
TOBII_REFRESH_RATE = 100  # Hz
TOBII_RESOLUTION = (1920, 1080)  # pixels
TOBII_WIDTH, TOBII_HEIGHT = 53.1, 30.0  # cm


## STRING CONSTANTS ##

####  General  ####
ID_STR = "id"
NAME_STR = "name"
TYPE_STR = "type"
COLOR_STR = "color"
URL_STR = "url"
ARTICLES_STR = "articles"
LICENSE_STR = "license"
OUTPUT_STR = "output"
WIDTH_STR, HEIGHT_STR = "width", "height"
RESOLUTION_STR = "resolution"
SAMPLING_RATE_STR = "sampling_rate"
PIXEL_STR = "pixel"
ITERATION_STR = "iteration"
DISTANCE_STR = "distance"
DIFFERENCE_STR = "difference"
THRESHOLD_STR = "threshold"
FIELD_NAME_STR = f"field_{NAME_STR}"
RUNTIME_STR = "runtime"

#### Trial Field Names  ####
SUBJECT_STR = "subject"
SUBJECT_ID_STR = f"{SUBJECT_STR}_{ID_STR}"
TRIAL_STR = "trial"
TRIAL_ID_STR = f"{TRIAL_STR}_{ID_STR}"
STIMULUS_STR = "stimulus"
STIMULUS_TYPE_STR = f"{STIMULUS_STR}_{TYPE_STR}"
STIMULUS_NAME_STR = f"{STIMULUS_STR}_{NAME_STR}"
CHANNEL_STR = "channel"
CHANNEL_TYPE_STR = f"{CHANNEL_STR}_{TYPE_STR}"

# TODO: replace these with enum
IMAGE_STR = "image"
VIDEO_STR = "video"
MOVING_DOT_STR = "moving_dot"

####  Eye Tracking Field Names  ####
TIME_STR = "time"
T, X, Y = "t", "x", "y"
PUPIL = "pupil"
LEFT_X, LEFT_Y, LEFT_PUPIL = f"left_{X}", f"left_{Y}", f"left_{PUPIL}"
RIGHT_X, RIGHT_Y, RIGHT_PUPIL = f"right_{X}", f"right_{Y}", f"right_{PUPIL}"

LABELER_STR = "labeler"
STATUS_STR = "status"
VIEWER_DISTANCE_STR = f"viewer_{DISTANCE_STR}"
PIXEL_SIZE_STR = f"{PIXEL_STR}_size"
IS_OUTLIER_STR = "is_outlier"
METADATA_STR = "metadata"

SAMPLE_STR, SAMPLES_STR = "sample", "samples"
LABEL_STR, LABELS_STR = "label", "labels"
EVENT_STR, EVENTS_STR = "event", "events"
MATCH_STR, MATCHES_STR = "match", "matches"
METRIC_STR, METRICS_STR = "metric", "metrics"

SAMPLES_CHANNEL_STR = f"{SAMPLES_STR}_{CHANNEL_STR}"
EVENTS_CHANNEL_STR = f"{EVENTS_STR}_{CHANNEL_STR}"

####  Event-Features  ####
START_STR, END_STR = "start", "end"
ONSET_STR, OFFSET_STR = "onset", "offset"
FEATURE_STR, FEATURES_STR = "feature", "features"
DURATION_STR = "duration"
VELOCITY_STR = "velocity"
ACCELERATION_STR = "acceleration"
AMPLITUDE_STR = "amplitude"
AZIMUTH_STR = "azimuth"
COUNT_STR = "count"
AREA_STR = "area"
MAX_DURATION_STR, MIN_DURATION_STR = f"max_{DURATION_STR}", f"min_{DURATION_STR}"
START_TIME_STR, END_TIME_STR = f"{START_STR}_{TIME_STR}", f"{END_STR}_{TIME_STR}"
CUMULATIVE_DISTANCE_STR = f"cumulative_{DISTANCE_STR}"
CUMULATIVE_AMPLITUDE_STR = f"cumulative_{AMPLITUDE_STR}"
PEAK_VELOCITY_STR = f"peak_{VELOCITY_STR}"
MEDIAN_VELOCITY_STR = f"median_{VELOCITY_STR}"
MIN_VELOCITY_STR = f"min_{VELOCITY_STR}"
CENTER_PIXEL_STR = f"center_{PIXEL_STR}"
PIXEL_STD_STR = f"{PIXEL_STR}_std"
DISPERSION_STR = "dispersion"
ELLIPSE_AREA_STR = f"ellipse_{AREA_STR}"

####  Metrics  ####
ACCURACY_STR = "accuracy"
BALANCED_ACCURACY_STR = "balanced_accuracy"
COHENS_KAPPA_STR = "cohen's_kappa"
MCC_STR = "mcc"
COMPLEMENT_NLD_STR = "complement_nld"
PRECISION_STR = "precision"
RECALL_STR = "recall"
F1_STR = "f1"
D_PRIME_STR = "d_prime"
CRITERION_STR = "criterion"
FALSE_ALARM_RATE_STR = "false_alarm_rate"
MATCH_RATIO_STR = f"{MATCH_STR}_ratio"



import pandas as pd

from peyes._DataModels.DatasetLoader import (
    Lund2013DatasetLoader, IRFDatasetLoader, HFCDatasetLoader, GazeComDatasetLoader
)


def load_dataset(dataset_name: str, directory: str = None, save: bool = False, verbose: bool = False) -> pd.DataFrame:
    """
    Loads the dataset from the specified directory. If the directory is not specified or the dataset is not found, it is
    downloaded from the internet. If `save` is True and the dataset was downloaded, it is saved to the specified
    directory.

    :param dataset_name: str; the name of the dataset to load
        Must be one of: "Lund2013", "IRF", "HFC", "GazeCom"
    :param directory: str; the directory to load the dataset from or save the dataset to
    :param save: bool; if True, the dataset is saved to the specified directory
    :param verbose: bool; if True, prints additional information during the loading process
    :return: a DataFrame containing the dataset
    :raises ValueError: if `save` is True and `directory` is not specified
    """
    dataset_name_lower = dataset_name.lower().strip()
    if save and not directory:
        raise ValueError("Directory must be specified to save the dataset")
    if dataset_name_lower == "lund2013":
        return Lund2013DatasetLoader.load(directory, save, verbose)
    if dataset_name_lower == "irf":
        return IRFDatasetLoader.load(directory, save, verbose)
    if dataset_name_lower == "hfc":
        return HFCDatasetLoader.load(directory, save, verbose)
    if dataset_name_lower == "gazecom":
        return GazeComDatasetLoader.load(directory, save, verbose)
    raise NotImplementedError(f"Unknown dataset: {dataset_name}")



import peyes._utils.constants as _cnst
from peyes._DataModels.Event import EventSequenceType as _EventSequenceType
from peyes._DataModels.EventLabelEnum import EventLabelEnum as _EventLabelEnum

from peyes.event_metrics._get_features import get_features as _features
from peyes.event_metrics._rates_and_transitions import event_rate as _event_rate
from peyes.event_metrics._get_features import features_by_labels
from peyes.event_metrics._rates_and_transitions import microsaccade_rate, microsaccade_ratio, transition_matrix


def start_times(events: _EventSequenceType):
    return _features(events, _cnst.START_TIME_STR, verbose=False)


def end_times(events: _EventSequenceType):
    return _features(events, _cnst.END_TIME_STR, verbose=False)


def durations(events: _EventSequenceType):
    return _features(events, _cnst.DURATION_STR, verbose=False)


def amplitudes(events: _EventSequenceType):
    return _features(events, _cnst.AMPLITUDE_STR, verbose=False)


def azimuths(events: _EventSequenceType):
    return _features(events, _cnst.AZIMUTH_STR, verbose=False)


def center_pixels(events: _EventSequenceType):
    return _features(events, _cnst.CENTER_PIXEL_STR, verbose=False)


def counts(events: _EventSequenceType):
    aggregated_features = features_by_labels(events)
    return aggregated_features[_cnst.COUNT_STR]


def saccade_rate(events: _EventSequenceType):
    return _event_rate(events, _EventLabelEnum.SACCADE)


def blink_rate(events: _EventSequenceType):
    return _event_rate(events, _EventLabelEnum.BLINK)



import pandas as pd
import plotly.io as pio
import plotly.express as px

import peyes
from analysis._default_values._helpers import *

pio.renderers.default = "browser"

######################
# Load Events Data

stim_trial_ids = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)

all_events = pd.read_pickle(os.path.join(PROCESSED_DATA_DIR, DATASET_NAME, "events.pkl"))
all_events = all_events.xs(1, level=peyes.constants.ITERATION_STR, axis=1)  # Keep only the first iteration
all_events = all_events.loc[:, all_events.columns.get_level_values(peyes.constants.TRIAL_ID_STR).isin(stim_trial_ids)]
all_events = all_events.dropna(axis=0, how="all")

######################
# Count Events by Labeler

all_labelers = all_events.columns.get_level_values(peyes.constants.LABELER_STR).unique()
events_by_labelers = {
    lblr: all_events.xs(lblr, level=peyes.constants.LABELER_STR, axis=1).stack().dropna() for lblr in all_labelers
}
event_counts_by_labelers = pd.DataFrame({
    lblr: evnts.apply(lambda e: e.label).value_counts() for lblr, evnts in events_by_labelers.items()
}).fillna(0).T.rename(columns=lambda l: peyes.parse_label(l).name)

del stim_trial_ids, all_events, all_labelers, events_by_labelers

######################
# Create Bar Plot

fig = px.bar(event_counts_by_labelers, title="Event Counts", text_auto=True)
fig.update_layout(
    title_x=0.5,
    xaxis_title=peyes.constants.LABELER_STR,
    yaxis_title=peyes.constants.COUNT_STR,
    width=1200, height=600,
    legend=dict(title="Event:", orientation="h", yanchor="top", y=1.07, xanchor="left", x=0.21),
)
fig.show()

######################
# Save Figure

peyes.visualize.save_figure(
    fig, "event_counts_per_labeler", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)



import time
from abc import ABC, abstractmethod
from typing import final, Dict
import logging

import remodnav
from overrides import override
from scipy.signal import savgol_filter

import peyes._DataModels.config as cnfg
from peyes._utils.vector_utils import *
from peyes._utils.pixel_utils import *
from peyes._utils.event_utils import calculate_sampling_rate, parse_label
from peyes._DataModels.EventLabelEnum import EventLabelEnum, EventLabelSequenceType


class BaseDetector(ABC):
    """
    Base class for gaze event detectors, objects that indicate the type of gaze event (fixation, saccade, blink) at each
    sample in the gaze data. All inherited classes must implement the `_detect_impl` method, which is the core of the
    gaze event detection process.

    Detection process:
    1. Detect blinks, including padding them if necessary
    2. Set x and y to nan where blinks are detected
    3. Detect gaze events (using the class-specific logic, implemented in `_detect_impl` method)
    4. Ignore chunks of gaze-events that are shorter than `min_event_duration`
    5. Merge chunks of the same type that are separated by less than `min_event_duration`

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    """
    _ARTICLES: List[str]

    _MINIMUM_SAMPLES_IN_EVENT = 2   # minimum number of samples in an event

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
    ):
        self._missing_value = missing_value
        self._min_event_duration = min_event_duration
        if self._min_event_duration <= 0:
            raise ValueError("Minimum event duration must be positive")
        self._pad_blinks_ms = pad_blinks_ms
        if self._pad_blinks_ms < 0:
            raise ValueError("Time to pad blinks must be non-negative")
        self._metadata = {}     # additional metadata
        self._sr = np.nan       # sampling rate calculated in the detect method
        self._name = name if name else self.__class__.__name__

    @classmethod
    @abstractmethod
    def get_default_params(cls) -> Dict[str, float]:
        raise NotImplementedError

    @final
    def detect(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> (EventLabelSequenceType, dict):
        start_time = time.time()
        if not np.isfinite(viewer_distance_cm) or viewer_distance_cm <= 0:
            raise ValueError("Viewer distance must be a positive finite number")
        if not np.isfinite(pixel_size_cm) or pixel_size_cm <= 0:
            raise ValueError("Pixel size must be a positive finite number")
        t, x, y = self._reshape_vectors(t, x, y)
        self._sr = calculate_sampling_rate(t)
        labels = np.full_like(t, EventLabelEnum.UNDEFINED, dtype=EventLabelEnum)
        is_blink = self._detect_blinks(x, y)
        # detect blinks and replace blink-samples with NaN
        labels[is_blink] = EventLabelEnum.BLINK
        x_copy, y_copy = copy.deepcopy(x), copy.deepcopy(y)
        x_copy[is_blink] = np.nan
        y_copy[is_blink] = np.nan
        labels = self._detect_impl(t, x_copy, y_copy, labels, viewer_distance_cm, pixel_size_cm)
        labels = merge_chunks(labels, self.min_event_samples)
        labels = reset_short_chunks(labels, self.min_event_samples, EventLabelEnum.UNDEFINED)
        labels = [parse_label(l) for l in labels]
        self._metadata.update({
            cnst.SAMPLING_RATE_STR: self.sr,
            cnst.PIXEL_SIZE_STR: pixel_size_cm,
            cnst.VIEWER_DISTANCE_STR: viewer_distance_cm,
            cnst.RUNTIME_STR: time.time() - start_time,
        })
        return labels, copy.deepcopy(self._metadata)

    @abstractmethod
    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        raise NotImplementedError

    @final
    @property
    def name(self) -> str:
        return self._name

    @final
    @name.setter
    def name(self, value: str):
        if not value:
            raise ValueError("Name must be a non-empty string")
        self._name = value

    @final
    @property
    def algorithm_name(self) -> str:
        return self.__class__.__name__.removesuffix("Detector").lower()

    @final
    @property
    def missing_value(self) -> float:
        return self._missing_value

    @final
    @property
    def sr(self) -> float:
        return self._sr

    @final
    @property
    def min_event_duration(self) -> float:
        return self._min_event_duration

    @final
    @property
    def min_event_samples(self) -> int:
        """
        Calculates the minimum number of samples in an event, based on the minimum event duration and the sampling rate.
        Returns the maximum between the calculated number of samples 2 (minimum number of samples in an event).
        """
        num_samples = self._calc_num_samples(self.min_event_duration, self.sr)
        return np.nanmax([num_samples, self._MINIMUM_SAMPLES_IN_EVENT])

    @final
    @property
    def pad_blinks_ms(self) -> float:
        return self._pad_blinks_ms

    @final
    @property
    def pad_blinks_samples(self) -> int:
        return self._calc_num_samples(self.pad_blinks_ms, self.sr)

    @classmethod
    @final
    def articles(cls) -> List[str]:
        """ List of articles to cite when using this detector """
        if not cls._ARTICLES:
            raise AttributeError(f"Class {cls.__name__} must implement class attribute `_{cnst.ARTICLES_STR}`")
        return cls._ARTICLES

    @classmethod
    @final
    def documentation(cls) -> str:
        """ Returns the documentation of the class """
        algorithm_name = f"Algorithm:\t{cls.__name__.removesuffix('Detector')}"
        articles = "Articles:\n" + "\n".join([f"- {a}" for a in cls.articles()])
        docstring = cls.__doc__ if cls.__doc__ else ""
        return f"{algorithm_name}\n{articles}\n{docstring}"

    def _detect_blinks(
            self,
            x: np.ndarray,
            y: np.ndarray,
    ) -> np.ndarray:
        """
        Detects blink candidates in the given data:
        1. Identifies samples where x or y are missing as blinks
        2. Ignores chunks of blinks that are shorter than the minimum event duration
        3. Merges consecutive blink chunks separated by less than the minimum event duration
        4. Pads the blink candidates by the amount in `self._dilate_nans_by`

        :param x: x-coordinates
        :param y: y-coordinates
        :return: boolean array indicating blink candidates
        """
        is_blink = np.zeros_like(x, dtype=bool)
        # identify samples where x or y are missing as blinks
        is_missing_x = np.isnan(x) if np.isnan(self.missing_value) else x == self.missing_value
        is_missing_y = np.isnan(y) if np.isnan(self.missing_value) else y == self.missing_value
        is_blink[is_missing_x | is_missing_y] = True
        # merge consecutive blinks if they are close enough to each other:
        is_blink = merge_chunks(is_blink, self.min_event_samples)
        # ignore blinks if they are shorter than the minimum event duration:
        is_blink = reset_short_chunks(is_blink, self.min_event_samples, False)
        # pad remaining blinks by the given amount:
        pad_samples = self.pad_blinks_samples
        if pad_samples == 0:
            return is_blink
        new_is_blink = is_blink.copy()      # avoid overwriting the original array
        for i, val in enumerate(is_blink):  # read the original array, write to the new one
            if val:
                start = max(0, i - pad_samples)
                end = min(len(is_blink), i + pad_samples)
                new_is_blink[start:end] = True
        return new_is_blink

    def _reshape_vectors(self, t: np.ndarray, x: np.ndarray, y: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
        if not is_one_dimensional(t):
            raise ValueError("`t` must be one-dimensional")
        if not is_one_dimensional(x):
            raise ValueError("`x` must be one-dimensional")
        if not is_one_dimensional(y):
            raise ValueError("`y` must be one-dimensional")
        t = t.reshape(-1)
        x = x.reshape(-1)
        y = y.reshape(-1)
        if len(t) != len(x) or len(t) != len(y) or len(x) != len(y):
            raise ValueError("t, x and y must have the same length")
        return t, x, y

    @staticmethod
    def _calc_num_samples(duration: float, sr: float) -> int:
        """
        :param duration: in ms
        :param sr: in Hz
        :return: int; number of samples
        """
        return round(duration * sr / cnst.MILLISECONDS_PER_SECOND)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class IGlobalThresholdDetector(ABC):

    @staticmethod
    def _get_global_threshold(threshold_deg: float, unit: str, vd: float, ps: float) -> float:
        unit = unit.lower().strip()
        if unit == "deg":
            return threshold_deg
        if unit == "rad":
            return np.deg2rad(threshold_deg)
        if unit == "px":
            return visual_angle_to_pixels(threshold_deg, vd, ps, use_radians=False)
        raise ValueError(f"Invalid unit: {unit}")


class IVTDetector(BaseDetector, IGlobalThresholdDetector):
    """
    Implements the I-VT (velocity threshold) gaze event detection algorithm, as described in:
        Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols.
        In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78).

    General algorithm:
    1. Calculate the angular velocity of the gaze data
    2. Identify saccade candidates as samples with angular velocity greater than the threshold
    3. Assume undefined (non-blink) samples are fixations

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    :param saccade_velocity_threshold: the threshold for angular velocity, in degrees per second. Default is 45 degrees
        per-second, as suggested in the paper "One algorithm to rule them all? An evaluation and discussion of ten eye
        movement event-detection algorithms" (2016), Andersson et al.
    """

    _ARTICLES = [
        "Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols. " +
        "In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78)",
    ]

    _DEFAULT_SACCADE_VELOCITY_THRESHOLD = 45  # deg/s
    _SACCADE_VELOCITY_THRESHOLD_STR = "saccade_velocity_threshold"

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
            saccade_velocity_threshold: float = _DEFAULT_SACCADE_VELOCITY_THRESHOLD,
    ):
        super().__init__(missing_value, min_event_duration, pad_blinks_ms, name)
        if saccade_velocity_threshold <= 0:
            raise ValueError("Saccade velocity threshold must be positive")
        self._saccade_velocity_threshold = saccade_velocity_threshold

    @classmethod
    def get_default_params(cls) -> Dict[str, float]:
        return {
            cls._SACCADE_VELOCITY_THRESHOLD_STR: cls._DEFAULT_SACCADE_VELOCITY_THRESHOLD,
        }

    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        if not np.isfinite(viewer_distance_cm) or viewer_distance_cm <= 0:
            raise ValueError("Viewer distance must be a positive finite number")
        if not np.isfinite(pixel_size_cm) or pixel_size_cm <= 0:
            raise ValueError("Pixel size must be a positive finite number")
        labels = np.asarray(copy.deepcopy(labels), dtype=EventLabelEnum)
        px_velocities = calculate_velocities(x, y, t)
        px_threshold = self._get_global_threshold(self.saccade_velocity_threshold_deg, "px", viewer_distance_cm,
                                                  pixel_size_cm)
        labels[(labels != EventLabelEnum.BLINK) & (px_velocities > px_threshold)] = EventLabelEnum.SACCADE
        labels[(labels != EventLabelEnum.BLINK) & (px_velocities <= px_threshold)] = EventLabelEnum.FIXATION
        self._metadata.update({
            f"{self._SACCADE_VELOCITY_THRESHOLD_STR}_deg": self.saccade_velocity_threshold_deg,
            f"{self._SACCADE_VELOCITY_THRESHOLD_STR}_px": px_threshold,
        })
        return labels

    @property
    def saccade_velocity_threshold_deg(self) -> float:
        return self._saccade_velocity_threshold


class IVVTDetector(IVTDetector):
    """
    Implements the I-VVT (dual velocity threshold) gaze event detection algorithm, which builds on the I-VT algorithm
    by adding a second velocity threshold for smooth pursuit detection.
    The I-VT algorithm is described in:
        Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols.
        In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78).
    The I-VVT algorithm is described in:
        Komogortsev, O. V., & Karpov, A. (2013). Automated classification and scoring of smooth pursuit eye movements in
        the presence of fixations and saccades. Behavior research methods, 45, 203-215.

    General algorithm:
    1. Calculate the angular velocity of the gaze data
    2. Identify saccade candidates as samples with angular velocity greater than the saccade threshold
    3. Identify smooth pursuit candidates as samples with intermediate angular velocity - lower than the saccade
        threshold and greater than the smooth pursuit threshold.
    4. Assume undefined (non-blink) samples are fixations

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    :param saccade_velocity_threshold: the threshold for angular velocity, in degrees per second. Default is 45 degrees
        per-second, as suggested in the paper "One algorithm to rule them all? An evaluation and discussion of ten eye
        movement event-detection algorithms" (2016), Andersson et al.
    :param smooth_pursuit_velocity_threshold: the threshold for angular velocity, in degrees per second, that separates
        smooth pursuit from fixations. Default is 26 degrees per-second, as used in Komogortsev & Karpov (2013).
    """

    _ARTICLES = [
        "Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols. " +
        "In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78)",
        "Komogortsev, O. V., & Karpov, A. (2013). Automated classification and scoring of smooth pursuit eye " +
        "movements in the presence of fixations and saccades. Behavior research methods, 45, 203-215."
    ]

    _DEFAULT_SMOOTH_PURSUIT_VELOCITY_THRESHOLD = 26  # deg/s
    _SMOOTH_PURSUIT_VELOCITY_THRESHOLD_STR = "smooth_pursuit_velocity_threshold"

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
            saccade_velocity_threshold: float = IVTDetector._DEFAULT_SACCADE_VELOCITY_THRESHOLD,
            smooth_pursuit_velocity_threshold: float = _DEFAULT_SMOOTH_PURSUIT_VELOCITY_THRESHOLD,
    ):
        super().__init__(missing_value, min_event_duration, pad_blinks_ms, name, saccade_velocity_threshold)
        if smooth_pursuit_velocity_threshold <= 0:
            raise ValueError("Smooth pursuit velocity threshold must be positive")
        self._smooth_pursuit_velocity_threshold = smooth_pursuit_velocity_threshold

    @classmethod
    def get_default_params(cls) -> Dict[str, float]:
        return {
            cls._SACCADE_VELOCITY_THRESHOLD_STR: cls._DEFAULT_SACCADE_VELOCITY_THRESHOLD,
            cls._SMOOTH_PURSUIT_VELOCITY_THRESHOLD_STR: cls._DEFAULT_SMOOTH_PURSUIT_VELOCITY_THRESHOLD
        }

    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        labels = super()._detect_impl(t, x, y, labels, viewer_distance_cm, pixel_size_cm)
        labels[labels == EventLabelEnum.FIXATION] = EventLabelEnum.UNDEFINED  # reset fixation labels
        px_velocities = calculate_velocities(x, y, t)
        px_threshold = self._get_global_threshold(self.smooth_pursuit_velocity_threshold_deg, "px", viewer_distance_cm,
                                                  pixel_size_cm)
        labels[(labels == EventLabelEnum.UNDEFINED) & (px_velocities > px_threshold)] = EventLabelEnum.SMOOTH_PURSUIT
        labels[(labels == EventLabelEnum.UNDEFINED) & (px_velocities <= px_threshold)] = EventLabelEnum.FIXATION
        self._metadata.update({
            f"{self._SMOOTH_PURSUIT_VELOCITY_THRESHOLD_STR}_deg": self.smooth_pursuit_velocity_threshold_deg,
            f"{self._SMOOTH_PURSUIT_VELOCITY_THRESHOLD_STR}_px": px_threshold,
        })
        return labels

    @property
    def smooth_pursuit_velocity_threshold_deg(self) -> float:
        return self._smooth_pursuit_velocity_threshold


class IDTDetector(BaseDetector, IGlobalThresholdDetector):
    """
    Implements the I-DT (dispersion threshold) gaze event detection algorithm, as described in:
        Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols.
        In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78).

    General algorithm:
    1. Initialize a window spanning the first `window_duration` milliseconds
    2. Calculate the dispersion of the gaze data in the window
    3. If the dispersion is below the threshold, label all samples in the window as fixation and expand the window by a
        single sample. Otherwise, label the current sample as saccade and start a new window in the next sample
    4. Repeat until the end of the gaze data

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    :param dispersion_threshold: the threshold for dispersion, in degrees. Default is 0.5 DVA, as used in the original
        paper by Salvucci and Goldberg (2000). In the  Andersson et al. (2016), the suggested threshold is 2.7 DVA.
    :param window_duration: the duration of the window in milliseconds. Default is the minimal fixation duration from
        the configuration file. The original Salvucci and Goldberg (2000) paper suggests a window duration of 100 ms.
    """

    _ARTICLES = [
        "Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols. " +
        "In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78)",
    ]

    _DEFAULT_DISPERSION_THRESHOLD = 0.5  # DVA
    _DEFAULT_WINDOW_DURATION = cnfg.EVENT_MAPPING[EventLabelEnum.FIXATION][cnst.MIN_DURATION_STR]  # ms; original paper suggests 100 ms
    __DISPERSION_THRESHOLD_STR = "dispersion_threshold"
    __DEFAULT_WINDOW_DURATION_STR = "window_duration"

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
            dispersion_threshold: float = _DEFAULT_DISPERSION_THRESHOLD,
            window_duration: float = _DEFAULT_WINDOW_DURATION,
    ):
        super().__init__(missing_value, min_event_duration, pad_blinks_ms, name)
        if dispersion_threshold <= 0:
            raise ValueError("Dispersion threshold must be positive")
        self._dispersion_threshold = dispersion_threshold
        self._window_duration = window_duration

    @classmethod
    def get_default_params(cls) -> Dict[str, float]:
        return {
            cls.__DISPERSION_THRESHOLD_STR: cls._DEFAULT_DISPERSION_THRESHOLD,
            cls.__DEFAULT_WINDOW_DURATION_STR: cls._DEFAULT_WINDOW_DURATION,
        }

    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        labels = np.asarray(copy.deepcopy(labels), dtype=EventLabelEnum)
        ws = self._calculate_window_size_samples(t)
        px_threshold = self._get_global_threshold(self.dispersion_threshold_deg, "px", viewer_distance_cm,
                                                  pixel_size_cm)
        start_idx, end_idx = 0, ws
        is_fixation = False
        while end_idx <= len(t):
            dispersion = line_dispersion(x[start_idx:end_idx], y[start_idx:end_idx])
            if dispersion < px_threshold:
                # label all samples in the window as fixation and expand window to the right
                is_fixation = True
                labels[start_idx: end_idx] = EventLabelEnum.FIXATION
                end_idx += 1
            elif is_fixation:
                # start new window in the end of the old one
                start_idx = end_idx - 1
                end_idx = start_idx + ws
                is_fixation = False
            else:
                # label current sample as saccade and start new window in the next sample
                labels[start_idx] = EventLabelEnum.SACCADE
                start_idx += 1
                end_idx += 1
        self._metadata.update({
            f"{self.__DISPERSION_THRESHOLD_STR}_deg": self.dispersion_threshold_deg,
            f"{self.__DISPERSION_THRESHOLD_STR}_px": px_threshold,
        })
        return labels

    @property
    def dispersion_threshold_deg(self) -> float:
        return self._dispersion_threshold

    @property
    def window_duration_ms(self) -> float:
        return self._window_duration

    def _calculate_window_size_samples(self, t: np.ndarray) -> int:
        ws = self._calc_num_samples(self.window_duration_ms, self.sr)
        if ws < 2:
            raise ValueError(f"window_duration={ws}ms is too short for the given sampling_rate={self.sr}Hz")
        if ws >= len(t):
            raise ValueError(f"window_duration={ws}ms is too long for the given input data")
        return ws

    @staticmethod
    def _calculate_dispersion_length_px(xs: np.ndarray, ys: np.ndarray) -> float:
        """ Calculates the dispersion length of the gaze points (px units) """
        return max(xs) - min(xs) + max(ys) - min(ys)


class IDVTDetector(IDTDetector, IVTDetector):
    """
    Implements the I-DVT (dispersion and velocity thresholds) gaze event detection algorithm, which combines the
    velocity threshold and dispersion threshold algorithms, to detect fixations, saccades and smooth pursuits.
    The I-VT algorithm is described in:
        Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols.
        In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78).
    The I-DT algorithm is described in:
        Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols.
        In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78).
    The I-DVT algorithm is described in:
        Komogortsev, O. V., & Karpov, A. (2013). Automated classification and scoring of smooth pursuit eye movements
        in the presence of fixations and saccades. Behavior research methods, 45, 203-215.

    General algorithm:
    1. Identify fixations using the dispersion-threshold algorithm:
        1.A. Initialize a window spanning the first `window_duration` milliseconds
        1.B. Calculate the dispersion of the gaze data in the window
        1.C. If the dispersion is below the threshold, label all samples in the window as fixation and expand the window
            by a single sample. Otherwise, label the current sample as saccade and start a new window in the next sample
        1.D. Repeat until the end of the gaze data
    2. Identify saccades using the velocity-threshold algorithm:
        2.A. Calculate the angular velocity of the gaze data
        2.B. Identify saccade candidates as samples with angular velocity greater than the saccade threshold
    3. The remaining samples have lower velocity than the saccade threshold, and higher dispersion than the fixation
        threshold. These are classified as smooth pursuits.

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    :param saccade_velocity_threshold: the threshold for angular velocity, in degrees per second. Default is 45 degrees
        per-second, as suggested in the paper "One algorithm to rule them all? An evaluation and discussion of ten eye
        movement event-detection algorithms" (2016), Andersson et al.
    :param dispersion_threshold: the threshold for dispersion, in degrees. Default is 2.0 DVA, as used in the original
        paper by Komogortsev & Karpov (2013). In the  Andersson et al. (2016), they suggested threshold is 2.7 DVA.
    :param window_duration: the duration of the window in milliseconds. Default is the minimal fixation duration from
        the configuration file. The original Komogortsev & Karpov (2013) paper suggest a threshold of 110-150 ms.
    """

    _ARTICLES = [
        "Salvucci, D. D., & Goldberg, J. H. (2000). Identifying fixations and saccades in eye-tracking protocols. " +
        "In Proceedings of the Symposium on Eye Tracking Research & Applications (pp. 71-78)",
        "Komogortsev, O. V., & Karpov, A. (2013). Automated classification and scoring of smooth pursuit eye " +
        "movements in the presence of fixations and saccades. Behavior research methods, 45, 203-215."
    ]

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
            dispersion_threshold: float = IDTDetector._DEFAULT_DISPERSION_THRESHOLD,
            window_duration: float = IDTDetector._DEFAULT_WINDOW_DURATION,
            saccade_velocity_threshold: float = IVTDetector._DEFAULT_SACCADE_VELOCITY_THRESHOLD,
    ):
        super(IDVTDetector, self).__init__(
            missing_value, min_event_duration, pad_blinks_ms, name, dispersion_threshold, window_duration
        )
        self._saccade_velocity_threshold = saccade_velocity_threshold

    @classmethod
    def get_default_params(cls) -> Dict[str, float]:
        return {
            **IDTDetector.get_default_params(),
            **IVTDetector.get_default_params(),
        }

    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        idt_labels = IDTDetector._detect_impl(
            self, t, x, y, labels, viewer_distance_cm, pixel_size_cm
        )
        is_fixation = idt_labels == EventLabelEnum.FIXATION
        ivt_labels = IVTDetector._detect_impl(
            self, t, x, y, labels, viewer_distance_cm, pixel_size_cm
        )
        is_saccade = (ivt_labels == EventLabelEnum.SACCADE) & ~is_fixation
        is_smooth_pursuit = ~is_fixation & ~is_saccade

        labels = np.asarray(copy.deepcopy(labels), dtype=EventLabelEnum)
        labels[(labels == EventLabelEnum.UNDEFINED) & is_fixation] = EventLabelEnum.FIXATION
        labels[(labels == EventLabelEnum.UNDEFINED) & is_saccade] = EventLabelEnum.SACCADE
        labels[(labels == EventLabelEnum.UNDEFINED) & is_smooth_pursuit] = EventLabelEnum.SMOOTH_PURSUIT
        return labels


class EngbertDetector(BaseDetector):
    """
    Implements the algorithm described by Engbert, Kliegl, and Mergenthaler in
        "Microsaccades uncover the orientation of covert attention" (2003)
        "Microsaccades are triggered by low retinal image slip" (2006)

    Implementation is based on the following repositories:
        - https://shorturl.at/lyBE2
        - https://shorturl.at/DHJZ6

    General algorithm:
        1. Calculate the velocity of the gaze data in both axes
        2. Calculate the median-based-standard-deviation of the velocity in each axis
        3. Calculate the saccade threshold as the multiple of the median-based-standard-deviation with the
            `lambda_param` constant
        4. Identify saccade candidates as samples with velocity greater than the noise threshold

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    :param lambda_param: the multiplication coefficient used for calculating saccade threshold. Default is 5, as
        suggested in the original paper by Engbert & Mergethaler (2006) (note Engbert & Klielgl 2003 use lambda=6).
    :param deriv_window_size: number of samples (including the middle sample) used to calculate the velocity, meaning
        the velocity at time t is the difference between the sum of the `ws//2` samples after and before t, divided by
        the window size. Default is 5, as suggested in the original paper
    """

    _ARTICLES = [
        "Engbert, R. & Kliegl, R. (2003). Microsaccades uncover the orientation of covert attention. Vision Research",
        "Engbert, R., Mergenthaler, K., & Purves, D. (Ed.). (2006). Microsaccades are triggered by low retinal " +
        "image slip. PNAS Proceedings of the National Academy of Sciences of the United States of America",
    ]

    _DEFAULT_LAMBDA_PARAM = 5               # standard deviation multiplier
    _DEFAULT_DERIVATION_WINDOW_SIZE = 5     # number of samples used to calculate axial velocity
    __THRESHOLD_VELOCITY_STR = "threshold_velocity"
    __LAMBDA_PARAM_STR = "lambda_param"
    __DERIV_WINDOW_SIZE_STR = "deriv_window_size"

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
            lambda_param: float = _DEFAULT_LAMBDA_PARAM,
            deriv_window_size: int = _DEFAULT_DERIVATION_WINDOW_SIZE,
    ):
        super().__init__(missing_value, min_event_duration, pad_blinks_ms, name)
        self._lambda_param = lambda_param
        if self._lambda_param <= 0:
            raise ValueError("Lambda parameter must be positive")
        self._deriv_window_size = deriv_window_size
        if self._deriv_window_size <= 0:
            raise ValueError("Derivation window size must be positive")

    @classmethod
    def get_default_params(cls) -> Dict[str, float]:
        return {
            cls.__LAMBDA_PARAM_STR: cls._DEFAULT_LAMBDA_PARAM,
            cls.__DERIV_WINDOW_SIZE_STR: cls._DEFAULT_DERIVATION_WINDOW_SIZE,
        }

    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        labels = np.asarray(copy.deepcopy(labels), dtype=EventLabelEnum)
        x_velocity = self._axial_velocities_px(x, self.sr, self.deriv_window_size)
        y_velocity = self._axial_velocities_px(y, self.sr, self.deriv_window_size)
        x_thresh = self._median_standard_deviation(x_velocity) * self.lambda_param
        y_thresh = self._median_standard_deviation(y_velocity) * self.lambda_param
        ellipse = (x_velocity / x_thresh) ** 2 + (y_velocity / y_thresh) ** 2
        labels[ellipse < 1] = EventLabelEnum.FIXATION
        labels[ellipse >= 1] = EventLabelEnum.SACCADE
        self._metadata.update({
            f"{cnst.X}_{self.__THRESHOLD_VELOCITY_STR}_pxs": x_thresh,
            f"{cnst.Y}_{self.__THRESHOLD_VELOCITY_STR}_px": y_thresh,
        })
        return labels

    @property
    def lambda_param(self) -> float:
        return self._lambda_param

    @property
    def deriv_window_size(self) -> int:
        return self._deriv_window_size

    @staticmethod
    def _axial_velocities_px(
            arr: np.ndarray, sr: float, window_size: int
    ) -> np.ndarray:
        """
        Calculates the velocity (px/sec) along a single axis, based on the algorithm described in the original paper:
        1. Sum values in a window of size window_size//2, *before* the current sample:
            sum_before = arr(t-1) + arr(t-2) + ... + arr(t-ws//2)
        2. Sum values in a window of size window_size//2, *after* the current sample:
            sum_after = arr(t+1) + arr(t+2) + ... + arr(t+ws//2)
        3. Calculate the difference of the signal at the current sample:
            diff = sum_after - sum_before
        4. Calculate the velocity as the difference multiplied by sampling rate and divided by the window size:
            velocity = diff * sr / ws
        5. The first and last ws//2 samples are filled with np.NaN

        :param arr: the array of values to calculate the velocity for
        :param sr: the sampling rate
        :param window_size: the window size

        :return: velocity array
        """
        assert sr > 0 and np.isfinite(sr), "Sampling rate must be a positive finite number"
        assert window_size > 0, "Window size must be positive"
        half_ws = window_size // 2 if window_size % 2 == 0 else window_size // 2 + 1
        velocities = np.full_like(arr, np.nan, dtype=float)
        for idx in range(half_ws, len(arr) - half_ws):
            sum_before = np.sum(arr[idx - half_ws:idx])
            sum_after = np.sum(arr[idx + 1:idx + half_ws + 1])
            diff = sum_after - sum_before
            velocities[idx] = diff * sr / window_size
        return velocities

    @staticmethod
    def _median_standard_deviation(arr) -> float:
        """ Calculates the median-based standard-deviation of the input array """
        squared_median = np.power(np.nanmedian(arr), 2)
        median_of_squares = np.nanmedian(np.power(arr, 2))
        sd = np.sqrt(median_of_squares - squared_median)
        return float(np.nanmax([sd, 1e-10]))  # avoid division by zero in case of no variance

    @override
    def _reshape_vectors(self, t: np.ndarray, x: np.ndarray, y: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
        t, x, y = super()._reshape_vectors(t, x, y)
        if len(x) < 2 * self.deriv_window_size:
            raise ValueError(f"Derivation window ({self.deriv_window_size} samples) is too long for the given data")
        return t, x, y


class NHDetector(BaseDetector):
    """
    Implements the algorithm described by Nyström & Holmqvist in
        Nyström, M., Holmqvist, K. An adaptive algorithm for fixation, saccade, and glissade detection in eyetracking
        data. Behavior Research Methods 42, 188–204 (2010).
    The code is based on the Matlab implementation available in https://github.com/dcnieho/NystromHolmqvist2010, which
    was developed for the following article:
        Niehorster, D. C., Siu, W. W., & Li, L. (2015). Manual tracking enhances smooth pursuit eye movements. Journal
        of vision, 15(15), 11-11.
    License: Unspecified

    General algorithm:
        1. Calculate angular velocity & acceleration
        2. Denoise the data using SAVGOL filter
        3. Saccade Detection:
            3a. Detect velocity peaks
            3b. Detect saccade onset and offset surrounding each peak
            3c. Ignore saccades that are too short
        4. PSO (Glissade) Detection:
            4a. Detect samples with velocity exceeding the PSO threshold, that shortly follow a saccade offset
            4b. Find PSO offset
        5. Fixation Detection:
            5a. Detect samples that are not part of a saccade, PSO or noise
            5b. Ignore fixations that are too short

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    :param filter_duration_ms: Savitzky-Golay filter's duration (ms)
    :param filter_polyorder: Savitzky-Golay filter's polynomial order
    :param saccade_max_velocity: maximum saccade velocity (deg/s)
    :param saccade_max_acceleration: maximum saccade acceleration (deg/s^2)
    :param min_saccade_duration: minimum saccade duration (ms)
    :param min_fixation_duration: minimum fixation duration (ms)
    :param max_pso_duration: maximum PSO duration (ms)
    :param alpha_param: weight of saccade onset threshold when detecting saccade offset
    :param ignore_short_peak_durations: if True, excludes sporadic instances where velocity is above the PT, when
        detecting saccade peaks
    :param allow_high_psos: if True, includes PSOs with maximum velocity exceeding saccades' peak threshold (PT),
        given that the PSO's max velocity is still lower than the preceding saccade's max velocity
    """

    _ARTICLES = [
        "Nyström, M., Holmqvist, K. An adaptive algorithm for fixation, saccade, and glissade detection in " +
        "eyetracking data. Behavior Research Methods 42, 188–204 (2010)",
        "Niehorster, D. C., Siu, W. W., & Li, L. (2015). Manual tracking enhances smooth pursuit eye movements. " +
        "Journal of vision, 15(15), 11-11",
    ]

    _DEFAULT_FILTER_DURATION_MS = 2 * cnfg.EVENT_MAPPING[EventLabelEnum.SACCADE][cnst.MIN_DURATION_STR]     # ms
    _DEFAULT_FILTER_POLYORDER = 2                                                                           # unitless
    _DEFAULT_SACCADE_MAX_VELOCITY = 1000                                                                    # deg/s
    _DEFAULT_SACCADE_MAX_ACCELERATION = 100000                                                              # deg/s^2
    _DEFAULT_MIN_SACCADE_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.SACCADE][cnst.MIN_DURATION_STR]    # ms
    _DEFAULT_MIN_FIXATION_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.FIXATION][cnst.MIN_DURATION_STR]  # ms
    _DEFAULT_MAX_PSO_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.PSO][cnst.MAX_DURATION_STR]            # ms
    _DEFAULT_ALPHA_PARAM = 0.7                                                                              # unitless

    __FILTER_DURATION_MS_STR = "filter_duration_ms"
    __FILTER_POLYORDER_STR = "filter_polyorder"
    __SACCADE_MAX_VELOCITY_STR = "saccade_max_velocity"
    __SACCADE_MAX_ACCELERATION_STR = "saccade_max_acceleration"
    __MIN_SACCADE_DURATION_STR = "min_saccade_duration"
    __MIN_FIXATION_DURATION_STR = "min_fixation_duration"
    __MAX_PSO_DURATION_STR = "max_pso_duration"
    __ALPHA_PARAM_STR = "alpha_param"
    __IGNORE_SHORT_PEAK_DURATIONS_STR = "ignore_short_peak_durations"
    __ALLOW_HIGH_PSOS_STR = "allow_high_psos"
    __SACCADE_PEAK_VELOCITY_THRESHOLD_STR = "saccade_peak_velocity_threshold"
    __SACCADE_ONSET_VELOCITY_THRESHOLD_STR = "saccade_onset_velocity_threshold"

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
            filter_duration_ms: float = _DEFAULT_FILTER_DURATION_MS,
            filter_polyorder: int = _DEFAULT_FILTER_POLYORDER,
            saccade_max_velocity: float = _DEFAULT_SACCADE_MAX_VELOCITY,
            saccade_max_acceleration: float = _DEFAULT_SACCADE_MAX_ACCELERATION,
            min_saccade_duration: float = _DEFAULT_MIN_SACCADE_DURATION_MS,
            min_fixation_duration: float = _DEFAULT_MIN_FIXATION_DURATION_MS,
            max_pso_duration: float = _DEFAULT_MAX_PSO_DURATION_MS,
            alpha_param: float = _DEFAULT_ALPHA_PARAM,
            ignore_short_peak_durations: bool = True,   # whether to exclude sporadic samples from the calculation (default is True)
            allow_high_psos: bool = True,
    ):
        """
        Initialize a new Nyström & Holmqvist gaze event detector
        :param filter_duration_ms: Savitzky-Golay filter's duration (ms)
        :param filter_polyorder: Savitzky-Golay filter's polynomial order
        :param saccade_max_velocity: maximum saccade velocity (deg/s)
        :param saccade_max_acceleration: maximum saccade acceleration (deg/s^2)
        :param min_saccade_duration: minimum saccade duration (ms)
        :param min_fixation_duration: minimum fixation duration (ms)
        :param max_pso_duration: maximum PSO duration (ms)
        :param alpha_param: weight of saccade onset threshold when detecting saccade offset
        :param ignore_short_peak_durations: if True, excludes sporadic instances where velocity is above the PT, when
            detecting saccade peaks
        :param allow_high_psos: if True, includes PSOs with maximum velocity exceeding saccades' peak threshold (PT),
            given that the PSO's max velocity is still lower than the preceding saccade's max velocity
        :param missing_value: the value that indicates missing data in the gaze data, default is np.NaN
        :param min_event_duration: minimum duration of a gaze event (ms) default is 5 ms
        :param pad_blinks_ms: padding duration for blinks (ms), default is 0 ms
        """
        super().__init__(missing_value, min_event_duration, pad_blinks_ms, name)
        self._filter_duration = filter_duration_ms
        if self._filter_duration <= 0:
            raise ValueError("Filter duration must be positive")
        self._filter_polyorder = filter_polyorder
        if self._filter_polyorder <= 0:
            raise ValueError("Filter polyorder must be positive")
        self._saccade_max_velocity = saccade_max_velocity
        if self._saccade_max_velocity <= 0:
            raise ValueError("Saccade max velocity must be positive")
        self._saccade_max_acceleration = saccade_max_acceleration
        if self._saccade_max_acceleration <= 0:
            raise ValueError("Saccade max acceleration must be positive")
        self._min_saccade_duration = max(min_saccade_duration, self._min_event_duration)
        if self._min_saccade_duration < 0:
            raise ValueError("Minimum saccade duration must be non-negative")
        self._min_fixation_duration = max(min_fixation_duration, self._min_event_duration)
        if self._min_fixation_duration < 0:
            raise ValueError("Minimum fixation duration must be non-negative")
        self._max_pso_duration = max(max_pso_duration, self._min_event_duration)
        if self._max_pso_duration < 0:
            raise ValueError("Maximum PSO duration must be non-negative")
        self._alpha_param = alpha_param
        if not 0 <= self._alpha_param <= 1:
            raise ValueError("Alpha parameter must be between 0 and 1")
        self._beta_param = 1 - alpha_param
        self._ignore_short_peak_durations = ignore_short_peak_durations
        self._allow_high_psos = allow_high_psos

    @classmethod
    def get_default_params(cls) -> Dict[str, float]:
        return {
            cls.__FILTER_DURATION_MS_STR: cls._DEFAULT_FILTER_DURATION_MS,
            cls.__FILTER_POLYORDER_STR: cls._DEFAULT_FILTER_POLYORDER,
            cls.__SACCADE_MAX_VELOCITY_STR: cls._DEFAULT_SACCADE_MAX_VELOCITY,
            cls.__SACCADE_MAX_ACCELERATION_STR: cls._DEFAULT_SACCADE_MAX_ACCELERATION,
            cls.__MIN_SACCADE_DURATION_STR: cls._DEFAULT_MIN_SACCADE_DURATION_MS,
            cls.__MIN_FIXATION_DURATION_STR: cls._DEFAULT_MIN_FIXATION_DURATION_MS,
            cls.__MAX_PSO_DURATION_STR: cls._DEFAULT_MAX_PSO_DURATION_MS,
            cls.__ALPHA_PARAM_STR: cls._DEFAULT_ALPHA_PARAM,
            cls.__IGNORE_SHORT_PEAK_DURATIONS_STR: True,
            cls.__ALLOW_HIGH_PSOS_STR: True,
        }

    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        # detect noise:
        v, a = self._velocities_and_accelerations(x, y, viewer_distance_cm, pixel_size_cm)  # deg/s, deg/s^2
        is_noise = self._detect_noise(v, a)

        # denoise the data:
        x_copy, y_copy, v_copy, a_copy = x.copy(), y.copy(), v.copy(), v.copy()
        x_copy[is_noise] = np.nan
        y_copy[is_noise] = np.nan
        v_copy[is_noise] = np.nan
        a_copy[is_noise] = np.nan

        # detect start & end indices of saccades and PSOs:
        peak_threshold, onset_threshold = self._calculate_saccade_thresholds(v_copy)  # PT, OnT velocity thresholds
        saccades_info = self._extract_saccade_info(
            v_copy, peak_threshold, onset_threshold
        )  # saccade id -> (onset-idx, peak-idx, offset-idx, offset-threshold-velocity)
        psos_info = self._detect_psos(v_copy, saccades_info)  # saccade id -> (start_idx, end_idx, pso_type)

        # save and return results
        labels = self._classify_samples(
            labels, is_noise, saccades_info, psos_info, self._allow_high_psos
        )
        self._metadata.update({
            f"{self.__SACCADE_PEAK_VELOCITY_THRESHOLD_STR}_deg": peak_threshold,
            f"{self.__SACCADE_PEAK_VELOCITY_THRESHOLD_STR}_px": visual_angle_to_pixels(
                peak_threshold, viewer_distance_cm, pixel_size_cm, use_radians=False, keep_sign=False
            ),
            f"{self.__SACCADE_ONSET_VELOCITY_THRESHOLD_STR}_deg": onset_threshold,
            f"{self.__SACCADE_ONSET_VELOCITY_THRESHOLD_STR}_px": visual_angle_to_pixels(
                onset_threshold, viewer_distance_cm, pixel_size_cm, use_radians=False, keep_sign=False
            ),
        })
        return labels

    @property
    def filter_duration_ms(self) -> float:
        return self._filter_duration

    @property
    def filter_polyorder(self) -> int:
        return self._filter_polyorder

    @property
    def saccade_max_velocity_deg(self) -> float:
        return self._saccade_max_velocity

    @property
    def saccade_max_acceleration_deg(self) -> float:
        return self._saccade_max_acceleration

    @property
    def alpha(self) -> float:
        return self._alpha_param

    @property
    def beta(self) -> float:
        return self._beta_param

    @property
    def ignore_short_peak_durations(self) -> bool:
        return self._ignore_short_peak_durations

    @property
    def is_high_psos_allowed(self) -> bool:
        return self._allow_high_psos

    @property
    def min_fixation_samples(self) -> int:
        return self._calc_num_samples(self._min_fixation_duration, self.sr)

    def _velocities_and_accelerations(self, x: np.ndarray, y: np.ndarray, vd, ps) -> (np.ndarray, np.ndarray):
        """
        Calculates the 1st and 2nd derivatives of the gaze data, using the Savitzky-Golay filter. Then calculates the
        velocity and acceleration from the derivatives.

        Note the original paper calculates the velocity and acceleration using:
            v = sr * sqrt[(x')^2 + (y')^2] * pixel-to-angle-constant
            a = sr * sqrt[(x'')^2 + (y'')^2] * pixel-to-angle-constant
        We use `delta=1/self.sr` when calculating the derivatives, to account for sampling time, so we don't need to
        multiply by sr when computing `v` and `a`. See more in the scipy documentation and in the following links:
            - https://stackoverflow.com/q/56168730/8543025
            - https://github.com/scipy/scipy/issues/9910

        :param x: 1D array of x coordinates
        :param y: 1D array of y coordinates
        :return: angular velocity and acceleration of each point
        """
        px_to_deg_constant = pixels_to_visual_angle(1, vd, ps, False)
        ws = self._calc_num_samples(self.filter_duration_ms, self.sr)
        if ws <= self.filter_polyorder:
            raise RuntimeError(
                f"Cannot compute {self.filter_polyorder}-order Savitzky-Golay filter with window of duration {ws}ms " +
                "and sampling rate of {self.sr}Hz"
            )
        # calculate angular velocity (deg/s): v = sqrt((x')^2 + (y')^2) * pixel-to-angle-constant:
        dx = savgol_filter(x, ws, self._filter_polyorder, deriv=1, delta=1/self.sr)
        dy = savgol_filter(y, ws, self._filter_polyorder, deriv=1, delta=1/self.sr)
        v = np.sqrt(dx ** 2 + dy ** 2) * px_to_deg_constant
        # calculate angular acceleration (deg/s^2): a = sqrt((x'')^2 + (y'')^2) * pixel-to-angle-constant:
        ddx = savgol_filter(x, ws, self._filter_polyorder, deriv=2, delta=1/self.sr)
        ddy = savgol_filter(y, ws, self._filter_polyorder, deriv=2, delta=1/self.sr)
        a = np.sqrt(ddx ** 2 + ddy ** 2) * px_to_deg_constant
        return v, a

    def _detect_noise(self, v: np.ndarray, a: np.ndarray) -> np.ndarray:
        """
        Detects noise in the gaze data based on the angular velocity and acceleration.
        :param v: angular velocity of the gaze data
        :param a: angular acceleration of the gaze data
        :return: boolean array indicating noise samples
        """
        is_noise = np.zeros(len(v), dtype=bool)
        is_noise[v > self.saccade_max_velocity_deg] = True
        is_noise[a > self.saccade_max_acceleration_deg] = True

        # expand noise periods to include surrounding samples up to the median overall velocity
        median_v = np.nanmedian(v)
        noise_idxs = np.where(is_noise)[0]
        for idx in noise_idxs:
            start, end = idx, idx
            while start > 0 and v[start] > median_v:
                start -= 1
            while end < len(v) and v[end] > median_v:
                end += 1
            is_noise[start:end] = True
        return is_noise

    def _calculate_saccade_thresholds(self, v: np.ndarray, max_iters: int = 100) -> (float, float):
        """
        Calculates the threshold velocities for saccade peaks (PT) and saccade onsets (OnT):
        1. Initialize PT as the maximal value (in range 75-300 deg/s) that has at least one sample with higher velocity.
        2. Iteratively update PT until convergence:
            a. Find samples with velocity below PT.
            b. If enforce_min_dur is True:
                - ignore samples that aren't part of a chunk of length >= min_fixation_samples.
                - chunks of sufficient length are shortened by min_saccade_samples // 3 samples at each edge, to avoid
                    contamination from saccades.
            c. Calculate mean & std of velocity below PT.
            d. Update PT = mean + 6 * std
        3. After convergence, calculate OnT as the mean velocity of samples below PT + 3 * std of velocity below PT.
        See more details in https://shorturl.at/wyCH7.

        :param v: angular velocities of the gaze data (deg/s)
        :param max_iters: maximum number of iterations (default is 100)
        :return:
            pt (float) - threshold velocity for detecting saccade peaks
            ont (float) - threshold velocity for detecting saccade onsets
        """
        # find the starting PT value, by making sure there are at least 1 peak with higher velocity
        start_pt_options = np.arange(300, 74, -25)
        is_v_above_pt = (v > start_pt_options[:, np.newaxis]).any(axis=1)
        pt = start_pt_options[np.argmin(is_v_above_pt)] if any(is_v_above_pt) else np.median(start_pt_options)
        # iteratively update PT value until convergence
        pt_prev = 0
        is_below_pt = v <= pt
        min_chunk_size = self.min_fixation_samples
        num_edge_sample_to_drop = self._calc_num_samples(self._min_saccade_duration // 3, self.sr)
        while abs(pt - pt_prev) > 1 and max_iters > 0:
            pt_prev = pt
            max_iters -= 1
            is_below_pt = v <= pt
            if self.ignore_short_peak_durations:
                # only consider samples that are part of a chunk of length >= min_fixation_samples:
                chunks_below_pt = [
                    ch for ch in get_chunk_indices(is_below_pt) if  is_below_pt[ch[0]] and len(ch) >= min_chunk_size
                ]
                # drop samples at the edges of each chunk to avoid contamination from saccades:
                chunks_below_pt = [ch[num_edge_sample_to_drop: -num_edge_sample_to_drop] for ch in chunks_below_pt]
                if len(chunks_below_pt) > 0:
                    # concatenate the chunks to get the final boolean array:
                    is_below_pt = np.concatenate(chunks_below_pt)
            mu = np.nanmean(v[is_below_pt])
            sigma = np.nanstd(v[is_below_pt])
            pt = mu + 6 * sigma
        if max_iters == 0:
            raise RuntimeError("Failed to converge on PT_1 value for saccade detection")
        ont = np.nanmean(v[is_below_pt]) + 3 * np.nanstd(v[is_below_pt])
        return pt, ont

    def _extract_saccade_info(
            self, v: np.ndarray, pt: float, ont: float, min_peak_samples: int = 2
    ) -> Dict[int, Tuple[int, int, int, float]]:
        """
        Detects saccades in the gaze data based on the angular velocity:
        1. Detect samples with velocity exceeding the saccade peak threshold (PT)
        2. Find the 1st sample preceding each peak with velocity below the onset threshold (OnT) and is a local minimum
        3. Find the 1st sample following each peak with velocity below the offset threshold (OfT) and is a local minimum
        4. Match each saccade peak-idx with its onset-idx, offset-idx and offset-threshold-velocity

        :param v: angular velocity of the gaze data
        :param pt: saccades' peak threshold velocity
        :param ont: saccades' onset threshold velocity
        :param min_peak_samples: minimum number of samples for a peak to be considered a saccade, otherwise ignored
        :return: dict mapping saccade id -> (onset-idx, peak-idx, offset-idx, offset-threshold-velocity)
        """
        is_above_pt = v > pt
        chunks_above_pt = [     # assume very short peaks are noise
            ch for ch in get_chunk_indices(is_above_pt) if is_above_pt[ch[0]] and len(ch) >= min_peak_samples
        ]
        # each chunk is a possible saccade, find the onset and offset of each saccade
        saccades_info = {}  # saccade_id -> (start_idx, peak_idx, end_idx, offset_threshold)
        for i, chunk in enumerate(chunks_above_pt):
            peak_idx: int = int(chunk[0])
            # find the onset of the saccade: the 1st local minimum preceding the peak with v<=OnT
            onset_idx = self.__find_local_minimum_index(v, peak_idx, ont, move_back=True)
            # calculate the offset threshold: OfT = a * OnT + b * OtT
            # note the locally adaptive term: OtT = mean(v) + 3 * std(v) for the min_fixation_samples before the onset
            window_start_idx = max(0, onset_idx - self.min_fixation_samples)
            window_vel = v[window_start_idx: onset_idx]
            ott = np.nanmean(window_vel) + 3 * np.nanstd(window_vel)
            if np.isfinite(ott) and ott < pt:
                offset_threshold = self.alpha * ont + self.beta * ott
            else:
                offset_threshold = ont
            # save saccade info
            last_peak_idx: int = int(chunk[-1])
            offset_idx = self.__find_local_minimum_index(v, last_peak_idx, offset_threshold, move_back=False)
            saccades_info[i] = (onset_idx, peak_idx, offset_idx, offset_threshold)
        return saccades_info

    def _detect_psos(
            self, v: np.ndarray, saccade_info: Dict[int, Tuple[int, int, int, float]]
    ) -> Dict[int, Tuple[int, int, bool]]:
        """
        Detects PSOs in the gaze data based on the angular velocity:
        1. Determine what velocity threshold to use for PSO detection (depending on value of self._detect_high_psos)
        2. Check if a window of length min fixation duration, succeeding each saccade, has samples above AND below the
              threshold. If so, there is a PSO.
        3. Identify the end of the PSO - the first local velocity-minimum after the last sample above the threshold
        4. Ignore PSOs with amplitude exceeding the preceding saccade

        :param v: angular velocity of the gaze data
        :param saccade_info: dictionary of saccade id -> (onset-idx, peak-idx, offset-idx, offset-threshold-velocity)
        :return: dict matching saccade id with PSO start-idx, end-idx and PSO type (high or low)
        """
        max_pso_samples = self._calc_num_samples(self._max_pso_duration, self.sr)
        pso_info = {}  # saccade_id -> (start_idx, end_idx, pso_type)

        # extract saccade indices to ready-to-use lists
        saccade_info_list = sorted(saccade_info.items(), key=lambda x: x[0])
        saccade_onset_idxs = np.array([info[1][0] for info in saccade_info_list])
        saccade_peak_idxs = np.array([info[1][1] for info in saccade_info_list])
        saccade_offset_idxs = np.array([info[1][2] for info in saccade_info_list])

        # find PSO start & end idxs after each saccade
        sac_id = 0
        while sac_id < len(saccade_info):
            sac_onset_idx, sac_peak_idx, sac_offset_idx, sac_offset_threshold = saccade_info[sac_id]
            # if a window succeeding a saccade has samples above AND below the offset threshold, there is a PSO
            start_idx, end_idx = 1 + sac_offset_idx, min([1 + sac_offset_idx + self.min_fixation_samples, len(v)])
            window = v[start_idx: end_idx]
            is_above, is_below = window > sac_offset_threshold, window < sac_offset_threshold
            if not (any(is_above) and any(is_below)):
                # no PSO for this saccade
                sac_id += 1
                continue

            # if the window contains samples with velocity above saccades' peak threshold, but below the preceding
            # saccade's max velocity, it is considered "high" PSO and contains saccade peaks
            end_idx = start_idx + np.where(is_above)[0][-1]
            is_high_pso = False
            is_peak_in_window = [start_idx <= p_idx < end_idx for p_idx in saccade_peak_idxs]
            if any(is_peak_in_window):
                last_peak = np.where(is_peak_in_window)[0][-1]
                last_offset_idx = saccade_offset_idxs[last_peak]
                if v[start_idx: last_offset_idx].max() < v[sac_onset_idx: sac_offset_idx].max():
                    # only allow high PSO if its max velocity is below the previous saccade's max velocity
                    is_high_pso = True
                    end_idx = max([end_idx, last_offset_idx])

            # move forward from the last sample above the threshold to find the first local minimum
            window = v[end_idx: len(v)]
            min_idx_in_window = self.__find_local_minimum_index(window, 0, sac_offset_threshold, move_back=False)
            end_idx += min_idx_in_window
            if end_idx - start_idx > max_pso_samples:
                # PSO is too long, ignore it
                sac_id += 1
                continue

            # save PSO info
            pso_info[sac_id] = (start_idx, end_idx, is_high_pso)

            # move to the next saccade
            next_sac_idx = np.where(saccade_onset_idxs > end_idx)[0]
            if len(next_sac_idx):
                sac_id = next_sac_idx[0]
            else:
                break
        return pso_info

    @staticmethod
    def _classify_samples(
            labels: np.ndarray,
            is_noise: np.ndarray,
            saccade_info: Dict[int, Tuple[int, int, int, float]],
            pso_info: Dict[int, Tuple[int, int, bool]],
            allow_high_psos: bool,
    ) -> (np.ndarray, np.ndarray):
        """
        Classifies each sample as either noise, saccade, PSO, fixation or blink. Samples that are not classified as
        noise, blink, saccade or PSO are considered fixations.

        If we allow high PSOs we override the saccade classification when a high PSO was also detected (i.e. we consider
        a saccade immediately following a previous saccade as high-PSO).
        Note the matlab implementation does the opposite, and first merges subsequent saccades that are only separated
        by a few PSO samples, and classifies the union as a saccade. We don't do this here (see in https://shorturl.at/EMOQR)

        :param is_noise: boolean array indicating noise samples
        :param saccade_info: dict of saccade id -> (onset-idx, peak-idx, offset-idx, offset-threshold-velocity)
        :param pso_info: dict of saccade -> (PSO start-idx, PSO end-idx and PSO type (high or low))
        :return: array of classified samples
        """
        labels = np.asarray(copy.deepcopy(labels), dtype=EventLabelEnum)
        for val in saccade_info.values():
            onset_idx, _, offset_idx, _ = val
            labels[onset_idx: offset_idx] = EventLabelEnum.SACCADE
        for val in pso_info.values():
            start_idx, end_idx, is_high = val
            if is_high and not allow_high_psos:
                # high PSO are essentially saccades that immediately follow a previous saccade
                continue
            labels[start_idx: end_idx] = EventLabelEnum.PSO

        is_blinks = labels == EventLabelEnum.BLINK
        is_saccade = labels == EventLabelEnum.SACCADE
        is_pso = labels == EventLabelEnum.PSO
        labels[~(is_noise | is_saccade | is_pso | is_blinks)] = EventLabelEnum.FIXATION
        return labels

    @staticmethod
    def __find_local_minimum_index(arr: np.ndarray, idx: int, min_thresh=np.inf, move_back=False) -> int:
        """
        Finds a local minimum in the array (an element that is smaller than its neighbors) starting from the given index.
        :param arr: the array to search in
        :param idx: the starting index
        :param min_thresh: the minimum value for a local minimum    (default: infinity)
        :param move_back: whether to move back or forward from the starting index   (default: False)
        :return: the index of the local minimum
        """
        while 0 < idx < len(arr):
            if arr[idx] < min_thresh and arr[idx] < arr[idx + 1] and arr[idx] < arr[idx - 1]:
                # idx is a local minimum
                return idx
            idx = idx - 1 if move_back else idx + 1
        return idx


class REMoDNaVDetector(BaseDetector):
    """
    This is a wrapper class that uses an implementation of the REMoDNaV algorithm to detect gaze events in the gaze
        data. This algorithm is based on the NHDetector algorithm, but extends and improves it by adding a more
        sophisticated saccade/pso detection algorithm, and by adding a smooth pursuit detection algorithm.

    See the REMoDNaV paper:
        Dar AH, Wagner AS, Hanke M. REMoDNaV: robust eye-movement classification for dynamic stimulation. Behav Res
        Methods. 2021 Feb;53(1):399-414. doi: 10.3758/s13428-020-01428-x
    See the NH Detector paper:
        Nyström, M., Holmqvist, K. An adaptive algorithm for fixation, saccade, and glissade detection in eyetracking
        data. Behavior Research Methods 42, 188–204 (2010).

    See the REMoDNaV algorithm documentation & implementation:
        https://github.com/psychoinformatics-de/remodnav/tree/master
    License: MIT

    :param missing_value: the value that indicates missing data in the gaze data.
    :param min_event_duration: the minimum duration of a gaze event, in milliseconds.
    :param pad_blinks_ms: the duration to pad around detected blinks, in milliseconds.
    :param name: the name of the detector. Default is the class name.
    :param median_filter_duration_ms: the duration of the median filter (ms), default is 50 ms
    :param savgol_filter_polyorder: the polynomial order for the Savitzky-Golay filter, default is 2
    :param savgol_filter_duration_ms: the duration of the Savitzky-Golay filter (ms), default is 19 ms
    :param max_velocity: the maximum velocity of the gaze data (deg/s), default is 1500
    :param min_saccade_duration: the minimum duration of a saccade (ms), default is 6 ms
    :param saccade_initial_velocity_threshold: the initial velocity threshold for saccade detection (deg/s), default is 300
    :param saccade_context_window_duration: the duration of the context window for saccade detection (ms), default is 1000
    :param saccade_initial_max_freq: the initial maximum frequency for saccade detection (Hz), default is 2.0
    :param saccade_onset_threshold_noise_factor: the noise factor for saccade onset threshold, default is 5.0
    :param min_smooth_pursuit_duration: the minimum duration of a smooth pursuit (ms), default is 40 ms
    :param smooth_pursuits_lowpass_cutoff_freq: the lowpass cutoff frequency for smooth pursuit detection (Hz), default is 4.0
    :param smooth_pursuit_drift_velocity_threshold: the drift velocity threshold for smooth pursuit detection (deg/s), default is 2.0
    :param min_fixation_duration: the minimum duration of a fixation (ms), default is 55 ms, used in the article by
        Andersson et al. (2017). The original REMoDNaV article (Dar et al. 2021) uses 50ms threshold.
    :param min_blink_duration: the minimum duration of a blink (ms), default is 20 ms
    :param max_pso_duration: the maximum duration of a PSO (ms), default is 80 ms. Original REMoDNaV article (Dar et al.
        2021) uses 40ms threshold.
    """

    _ARTICLES = [
        "Dar AH, Wagner AS, Hanke M. REMoDNaV: robust eye-movement classification for dynamic stimulation. Behav " +
        "Res Methods. 2021 Feb;53(1):399-414. doi: 10.3758/s13428-020-01428-x",
    ]

    ## Default Parameters

    _DEFAULT_MEDIAN_FILTER_DURATION_MS = 50                 # ms
    _DEFAULT_SAVGOL_POLYORDER = 2                           # unitless
    _DEFAULT_SAVGOL_DURATION_MS = 19                        # ms
    _DEFAULT_MAX_VELOCITY_DEG = 1500                        # deg/s

    _DEFAULT_MIN_FIXATION_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.FIXATION][cnst.MIN_DURATION_STR]
    _DEFAULT_MIN_BLINK_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.BLINK][cnst.MIN_DURATION_STR]
    _DEFAULT_MAX_PSO_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.PSO][cnst.MAX_DURATION_STR]

    _DEFAULT_MIN_SACCADE_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.SACCADE][cnst.MIN_DURATION_STR]
    _DEFAULT_SACCADE_INITIAL_VELOCITY_THRESHOLD = 300       # deg/s
    _DEFAULT_SACCADE_CONTEXT_WINDOW_DURATION_MS = 1000      # ms
    _DEFAULT_SACCADE_INITIAL_MAX_FREQ = 2.0                 # Hz
    _DEFAULT_SACCADE_ONSET_THRESHOLD_NOISE_FACTOR = 5.0     # unitless

    _DEFAULT_MIN_SMOOTH_PURSUIT_DURATION_MS = cnfg.EVENT_MAPPING[EventLabelEnum.SMOOTH_PURSUIT][cnst.MIN_DURATION_STR]
    _DEFAULT_SMOOTH_PURSUIT_LOWPASS_CUTOFF_FREQ = 4.0       # Hz
    _DEFAULT_SMOOTH_PURSUIT_DRIFT_VELOCITY_THRESHOLD = 2.0  # deg/s

    __LABEL_MAPPING = {
        'FIXA': EventLabelEnum.FIXATION, 'SACC': EventLabelEnum.SACCADE, 'ISAC': EventLabelEnum.SACCADE,
        'HPSO': EventLabelEnum.PSO, 'IHPS': EventLabelEnum.PSO, 'LPSO': EventLabelEnum.PSO,
        'ILPS': EventLabelEnum.PSO, 'PURS': EventLabelEnum.SMOOTH_PURSUIT, 'BLNK': EventLabelEnum.BLINK
    }

    ## Parameter Names

    __MEDIAN_FILTER_DURATION_MS_STR = "median_filter_duration_ms"
    __SAVGOL_POLYORDER_STR = "savgol_filter_polyorder"
    __SAVGOL_DURATION_MS_STR = "savgol_filter_duration_ms"
    __MAX_VELOCITY_STR = "max_velocity"

    __MIN_FIXATION_DURATION_STR = "min_fixation_duration"
    __MIN_BLINK_DURATION_STR = "min_blink_duration"
    __MAX_PSO_DURATION_STR = "max_pso_duration"

    __MIN_SACCADE_DURATION_STR = "min_saccade_duration"
    __SACCADE_INITIAL_VELOCITY_THRESHOLD_STR = "saccade_initial_velocity_threshold"
    __SACCADE_CONTEXT_WINDOW_DURATION_STR = "saccade_context_window_duration"
    __SACCADE_INITIAL_MAX_FREQ_STR = "saccade_initial_max_freq"
    __SACCADE_ONSET_THRESHOLD_NOISE_FACTOR_STR = "saccade_onset_threshold_noise_factor"

    __MIN_SMOOTH_PURSUIT_DURATION_STR = "min_smooth_pursuit_duration"
    __SMOOTH_PURSUIT_LOWPASS_CUTOFF_FREQ_STR = "smooth_pursuits_lowpass_cutoff_freq"
    __SMOOTH_PURSUIT_DRIFT_VELOCITY_THRESHOLD_STR = "smooth_pursuit_drift_velocity_threshold"
    __SHOW_WARNINGS_STR = "show_warnings"

    def __init__(
            self,
            missing_value: float,
            min_event_duration: float,
            pad_blinks_ms: float,
            name: str = None,
            median_filter_duration_ms: float = _DEFAULT_MEDIAN_FILTER_DURATION_MS,
            savgol_filter_polyorder: int = _DEFAULT_SAVGOL_POLYORDER,
            savgol_filter_duration_ms: float = _DEFAULT_SAVGOL_DURATION_MS,
            max_velocity: float = _DEFAULT_MAX_VELOCITY_DEG,
            min_saccade_duration: float = _DEFAULT_MIN_SACCADE_DURATION_MS,
            saccade_initial_velocity_threshold: float = _DEFAULT_SACCADE_INITIAL_VELOCITY_THRESHOLD,
            saccade_context_window_duration: float = _DEFAULT_SACCADE_CONTEXT_WINDOW_DURATION_MS,
            saccade_initial_max_freq: float = _DEFAULT_SACCADE_INITIAL_MAX_FREQ,
            saccade_onset_threshold_noise_factor: float = _DEFAULT_SACCADE_ONSET_THRESHOLD_NOISE_FACTOR,
            min_smooth_pursuit_duration: float = _DEFAULT_MIN_SMOOTH_PURSUIT_DURATION_MS,
            smooth_pursuits_lowpass_cutoff_freq: float = _DEFAULT_SMOOTH_PURSUIT_LOWPASS_CUTOFF_FREQ,
            smooth_pursuit_drift_velocity_threshold: float = _DEFAULT_SMOOTH_PURSUIT_DRIFT_VELOCITY_THRESHOLD,
            min_fixation_duration: float = _DEFAULT_MIN_FIXATION_DURATION_MS,
            min_blink_duration: float = _DEFAULT_MIN_BLINK_DURATION_MS,
            max_pso_duration: float = _DEFAULT_MAX_PSO_DURATION_MS,
            show_warnings: bool = True,
    ):
        super().__init__(missing_value, min_event_duration, pad_blinks_ms, name)
        self._median_filter_length = median_filter_duration_ms
        self._min_saccade_duration_ms = max(min_saccade_duration, self._min_event_duration)
        self._saccade_initial_velocity_threshold = saccade_initial_velocity_threshold
        self._saccade_context_window_duration = saccade_context_window_duration
        self._saccade_initial_max_freq = saccade_initial_max_freq
        self._saccade_onset_threshold_noise_factor = saccade_onset_threshold_noise_factor
        self._min_smooth_pursuit_duration_ms = max(min_smooth_pursuit_duration, self._min_event_duration)
        self._smooth_pursuit_lowpass_cutoff_freq = smooth_pursuits_lowpass_cutoff_freq
        self._smooth_pursuit_drift_velocity_threshold = smooth_pursuit_drift_velocity_threshold
        self._min_fixation_duration_ms = max(min_fixation_duration, self._min_event_duration)
        self._min_blink_duration_ms = max(min_blink_duration, self._min_event_duration)
        self._max_pso_duration_ms = max(max_pso_duration, self._min_event_duration)
        self._savgol_polyorder = savgol_filter_polyorder
        self._savgol_duration_ms = savgol_filter_duration_ms
        self._max_velocity = max_velocity
        self.show_warnings = show_warnings

    @classmethod
    def get_default_params(cls) -> Dict[str, float]:
        return {
            cls.__MIN_SACCADE_DURATION_STR: cls._DEFAULT_MIN_SACCADE_DURATION_MS,
            cls.__SACCADE_INITIAL_VELOCITY_THRESHOLD_STR: cls._DEFAULT_SACCADE_INITIAL_VELOCITY_THRESHOLD,
            cls.__SACCADE_CONTEXT_WINDOW_DURATION_STR: cls._DEFAULT_SACCADE_CONTEXT_WINDOW_DURATION_MS,
            cls.__SACCADE_INITIAL_MAX_FREQ_STR: cls._DEFAULT_SACCADE_INITIAL_MAX_FREQ,
            cls.__SACCADE_ONSET_THRESHOLD_NOISE_FACTOR_STR: cls._DEFAULT_SACCADE_ONSET_THRESHOLD_NOISE_FACTOR,
            cls.__MIN_SMOOTH_PURSUIT_DURATION_STR: cls._DEFAULT_MIN_SMOOTH_PURSUIT_DURATION_MS,
            cls.__SMOOTH_PURSUIT_LOWPASS_CUTOFF_FREQ_STR: cls._DEFAULT_SMOOTH_PURSUIT_LOWPASS_CUTOFF_FREQ,
            cls.__SMOOTH_PURSUIT_DRIFT_VELOCITY_THRESHOLD_STR: cls._DEFAULT_SMOOTH_PURSUIT_DRIFT_VELOCITY_THRESHOLD,
            cls.__MIN_FIXATION_DURATION_STR: cls._DEFAULT_MIN_FIXATION_DURATION_MS,
            cls.__MIN_BLINK_DURATION_STR: cls._DEFAULT_MIN_BLINK_DURATION_MS,
            cls.__MAX_PSO_DURATION_STR: cls._DEFAULT_MAX_PSO_DURATION_MS,
            cls.__SAVGOL_POLYORDER_STR: cls._DEFAULT_SAVGOL_POLYORDER,
            cls.__SAVGOL_DURATION_MS_STR: cls._DEFAULT_SAVGOL_DURATION_MS,
            cls.__MEDIAN_FILTER_DURATION_MS_STR: cls._DEFAULT_MEDIAN_FILTER_DURATION_MS,
            cls.__MAX_VELOCITY_STR: cls._DEFAULT_MAX_VELOCITY_DEG,
            cls.__SHOW_WARNINGS_STR: True,
        }

    def _detect_impl(
            self,
            t: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            labels: np.ndarray,
            viewer_distance_cm: float,
            pixel_size_cm: float,
    ) -> np.ndarray:
        if not self.show_warnings:
            lgr = logging.getLogger('remodnav.clf')
            lgr.setLevel(logging.WARNING + 1)   # ignore warnings from the REMoDNaV library
        classifier = remodnav.EyegazeClassifier(
            px2deg=pixels_to_visual_angle(1, viewer_distance_cm, pixel_size_cm, use_radians=False),
            sampling_rate=self.sr,
            min_saccade_duration=self._min_saccade_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            min_intersaccade_duration=self.min_inter_saccade_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            saccade_context_window_length=self.saccade_context_window_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            velthresh_startvelocity=self.saccade_initial_velocity_threshold,
            max_initial_saccade_freq=self.saccade_initial_max_freq,
            noise_factor=self.saccade_onset_threshold_noise_factor,
            min_pursuit_duration=self.min_smooth_pursuit_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            pursuit_velthresh=self.smooth_pursuit_drift_velocity_threshold,
            lowpass_cutoff_freq=self.smooth_pursuits_lowpass_cutoff_freq,
            min_fixation_duration=self.min_fixation_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            max_pso_duration=self.max_pso_duration_ms / cnst.MILLISECONDS_PER_SECOND,
        )
        xy = np.rec.fromarrays([x, y], names="{},{}".format(cnst.X, cnst.Y), formats="<f8,<f8")
        pp = classifier.preproc(
            xy,
            max_vel=self.max_velocity,
            savgol_polyord=self.savgol_filter_polyorder,
            savgol_length=self.savgol_filter_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            median_filter_length=self.median_filter_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            min_blink_duration=self.min_blink_duration_ms / cnst.MILLISECONDS_PER_SECOND,
            dilate_nan=self.pad_blinks_ms / cnst.MILLISECONDS_PER_SECOND,
        )
        detected_events = classifier(pp, classify_isp=True, sort_events=True)   # returns a list of dicts, each dict is a single gaze event
        labels = np.asarray(copy.deepcopy(labels), dtype=EventLabelEnum)
        for i, event in enumerate(detected_events):
            start_sample = round(event["start_time"] * self.sr)
            end_sample = round(event["end_time"] * self.sr)
            label = self.__LABEL_MAPPING[event["label"]]
            labels[start_sample:end_sample+1] = label
        return labels

    @property
    def min_saccade_duration_ms(self) -> float:
        return self._min_saccade_duration_ms

    @property
    def min_inter_saccade_duration_ms(self) -> float:
        return min(self._min_blink_duration_ms, self._min_fixation_duration_ms, self._min_smooth_pursuit_duration_ms)

    @property
    def saccade_initial_velocity_threshold(self) -> float:
        return self._saccade_initial_velocity_threshold  # deg/s

    @property
    def saccade_context_window_duration_ms(self) -> float:
        return self._saccade_context_window_duration

    @property
    def saccade_initial_max_freq(self) -> float:
        return self._saccade_initial_max_freq

    @property
    def saccade_onset_threshold_noise_factor(self) -> float:
        return self._saccade_onset_threshold_noise_factor

    @property
    def min_smooth_pursuit_duration_ms(self) -> float:
        return self._min_smooth_pursuit_duration_ms

    @property
    def smooth_pursuits_lowpass_cutoff_freq(self) -> float:
        return self._smooth_pursuit_lowpass_cutoff_freq

    @property
    def smooth_pursuit_drift_velocity_threshold(self) -> float:
        return self._smooth_pursuit_drift_velocity_threshold

    @property
    def min_fixation_duration_ms(self) -> float:
        return self._min_fixation_duration_ms

    @property
    def min_blink_duration_ms(self) -> float:
        return self._min_blink_duration_ms

    @property
    def max_pso_duration_ms(self) -> float:
        return self._max_pso_duration_ms

    @property
    def savgol_filter_polyorder(self) -> int:
        return self._savgol_polyorder

    @property
    def savgol_filter_duration_ms(self) -> float:
        return self._savgol_duration_ms

    @property
    def median_filter_duration_ms(self) -> float:
        return self._median_filter_length

    @property
    def max_velocity(self) -> float:
        return self._max_velocity   # deg/s



import os

import numpy as np
import peyes

import analysis.utils as u

DATASET_NAME = "lund2013"
STIMULUS_TYPE = peyes.constants.IMAGE_STR
GT1, GT2 = "RA", "MN"
MULTI_COMP = "bonferroni"   # method for multiple comparisons correction: bonferroni, fdr_bh, holm, etc.
ALPHA = 0.05

_PARAMS_EXPLORATION_STR = "parameter_exploration"
_FIGURES_STR = "figures"
PROCESSED_DATA_DIR = os.path.join(u.OUTPUT_DIR, _PARAMS_EXPLORATION_STR)
FIGURES_DIR = os.path.join(u.OUTPUT_DIR, _PARAMS_EXPLORATION_STR, DATASET_NAME, _FIGURES_STR)
os.makedirs(FIGURES_DIR, exist_ok=True)

_default_detector_params = dict(missing_value=np.nan, min_event_duration=4, pad_blinks_time=0)
DETECTORS = {
    "ivt_andersson": peyes.create_detector(     # Andersson et al. (2017)
        algorithm="ivt", saccade_velocity_threshold=45, **_default_detector_params
    ),
    "ivt_birawo": peyes.create_detector(        # Birawo & Kasprowski (2022)
        algorithm="ivt", saccade_velocity_threshold=16.5, **_default_detector_params
    ),
    # "ivt_salvucci": peyes.create_detector(    # Salvucci & Goldberg (2000)
    #     algorithm="ivt", saccade_velocity_threshold=100, **_default_detector_params
    # ),
    "ivvt": peyes.create_detector(
        algorithm="ivvt", saccade_velocity_threshold=45, smooth_pursuit_velocity_threshold=26, **_default_detector_params
    ),
    "idt_andersson": peyes.create_detector(
        algorithm="idt", dispersion_threshold=2.7, window_duration=55, **_default_detector_params
    ),
    # "idt_birawo": peyes.create_detector(
    #         # Birawo & Kasprowski (2022) use dispersion of 3.5px and window of 5 samples (10ms)
    #     algorithm="idt", dispersion_threshold=0.1, window_duration=10, **_default_detector_params
    # ),
    "idt_salvucci": peyes.create_detector(
        algorithm="idt", dispersion_threshold=1, window_duration=100, **_default_detector_params
    ),
    "idvt_andersson": peyes.create_detector(
        algorithm="idvt",
        saccade_velocity_threshold=45,
        dispersion_threshold=2.7,
        window_duration=55,
        **_default_detector_params
    ),
    "idvt_komogortsev": peyes.create_detector(
            # Komogortsev & Karpov (2013)
        algorithm="idvt",
        saccade_velocity_threshold=45,
        dispersion_threshold=2,
        window_duration=110,
        **_default_detector_params
    ),
    "idvt_komogortsev_birawo": peyes.create_detector(
            # Komogortsev & Karpov (2013) + Birawo & Kasprowski (2022)
        algorithm="idvt",
        saccade_velocity_threshold=16.5,
        dispersion_threshold=2,
        window_duration=110,
        **_default_detector_params
    ),
    "engbert_5": peyes.create_detector(
        algorithm="engbert", lambda_param=5, **_default_detector_params
    ),
    "engbert_6": peyes.create_detector(
        algorithm="engbert", lambda_param=6, **_default_detector_params
    ),
}
for key, detector in DETECTORS.items():
    detector.name = key




from peyes.alignment_metrics._timing_differences import onset_differences, offset_differences
from peyes.alignment_metrics._signal_detection_metrics import onset_detection_metrics, offset_detection_metrics



import warnings
from typing import Union, Sequence

import pandas as pd
import plotly.graph_objects as go
import plotly.express as px

import peyes
import peyes._utils.constants as cnst
import peyes._utils.visualization_utils as vis_utils
from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.EventLabelEnum import EventLabelEnum

from peyes import summarize_events


def feature_comparison(
        features: Union[str, Sequence[str]],
        *event_sequences: EventSequenceType,
        **kwargs
) -> go.Figure:
    """
    Creates a Ridge Plot comparing the distribution of features across multiple event sequences.
    Each feature is shown in a separate subplot.

    :param features: name(s) of the feature(s) to compare.
    :param event_sequences: array-like of Event objects

    :keyword include_outliers: bool; whether to include outliers in the plot (default is False)
    :keyword labels: array-like of str; labels for each event sequence (default is 1, 2, ...)
    :keyword colors: dict; color map for each label (default is the colormap from `utils.visualization_utils.py`)
    :keyword opacity: float; opacity of the violins (default is 0.75)
    :keyword line_width: float; width of the violin lines (default is 2)
    :keyword show_box: bool; whether to show the box plot (default is True)
    :keyword title: str; title for the plot (default is "Feature Comparison")

    :return: the plotly figure
    """
    include_outliers = kwargs.get("include_outliers", False)
    labels = kwargs.get("labels", list(range(len(event_sequences))))
    colors = vis_utils.get_label_colormap(kwargs.get("colors", None))
    if isinstance(features, str):
        features = [features]
    fig, nrows, ncols = vis_utils.make_empty_figure(
        list(map(lambda feat: feat.strip().replace("_", " ").title(), features)), sharex=False, sharey=True
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for i, (seq_name, ev_seq) in enumerate(zip(labels, event_sequences)):
            summary_df = summarize_events(ev_seq)
            if not include_outliers:
                summary_df = summary_df[~summary_df["is_outlier"]]
            for j, feat in enumerate(features):
                color = (
                        colors.get(seq_name, None) or colors.get(seq_name.strip().lower(), None) or
                        colors.get(seq_name.strip().upper(), None) or colors[i]
                )
                color = f"rgb{color}"
                r, c = (j, 0) if ncols == 1 else divmod(j, ncols)
                if feat.lower().strip() == peyes.constants.COUNT_STR:
                    num_events = summary_df.shape[0]
                    fig.add_trace(
                        row=r+1, col=c+1, trace=go.Bar(
                            x=[num_events], y=[seq_name], name=seq_name, legendgroup=seq_name,
                            marker=dict(color=color), showlegend=j == 0, orientation='h',
                        )
                    )
                else:
                    fig.add_trace(
                        row=r + 1, col=c + 1, trace=go.Violin(
                            x=summary_df[feat].dropna().values,
                            name=seq_name, legendgroup=seq_name, scalegroup=feat,
                            line_color=color, opacity=kwargs.get("opacity", 0.75), width=kwargs.get("line_width", 2),
                            box_visible=kwargs.get("show_box", True), points=False,
                            orientation='h', spanmode='hard', side='positive',
                            showlegend=j == 0,
                        )
                    )
        fig.update_layout(
            title=kwargs.get("title", "Feature Comparison"),
        )
    return fig


def main_sequence(
        saccades: EventSequenceType,
        y_feature: str = cnst.DURATION_STR,
        include_outliers: bool = True,
) -> (go.Figure, pd.DataFrame):
    """
    Creates a saccades "Main Sequence" plot, showing the relationship between saccade amplitude and either duration or
    peak velocity.

    :param saccades: array-like of saccade Event objects
    :param y_feature: str; the feature to plot on the y-axis. Must be either "duration" or "peak_velocity"
    :param include_outliers: bool; whether to include outliers in the plot (default is True)

    :return:
        - fig: go.Figure; the figure object
        - stat_results: pd.DataFrame; statistics of the inlier (and outlier) trendline(s)
    """
    saccades = [s for s in saccades if s.label == EventLabelEnum.SACCADE]
    if y_feature.lower() == cnst.DURATION_STR.lower():
        params = dict(
            title="Main Sequence: Duration vs. Amplitude",
            trendline_log_x=False, x_feature_units="DVA", y_feature_units="ms",
        )
    elif y_feature.lower() == cnst.PEAK_VELOCITY_STR.lower():
        params = dict(
            title="Main Sequence: Peak Velocity vs. Amplitude",
            trendline_log_x=True, x_feature_units="DVA", y_feature_units="DVA/s",
        )
    else:
        raise KeyError(f"Invalid `y_feature` argument: {y_feature}. Must be `{cnst.DURATION_STR}` or `{cnst.PEAK_VELOCITY_STR}`.")
    fig, stat_results = feature_relationship(
        events=saccades, include_outliers=include_outliers, trendline="trace" if include_outliers else "overall",
        x_feature="amplitude", y_feature=y_feature.lower(),
        marginal_x='box', marginal_y='box', **params
    )
    return fig, stat_results


def feature_relationship(
        events: EventSequenceType, x_feature: str, y_feature: str, title: str = None, **kwargs
) -> (go.Figure, pd.DataFrame):
    """
    Creates a scatter plot of the two features of the provided events, with optional trendline(s).

    :param events: array-like of Event objects
    :param x_feature: str; name of the feature to plot on the x-axis
    :param y_feature: str; name of the feature to plot on the y-axis
    :param title: str; title of the plot. if None, the default title is "`y_feature` vs. `x_feature`"

    :keyword include_outliers: bool; whether to include outliers in the plot (default is True)
    :keyword log_x: bool; whether to use a log scale on the x-axis (default is False)
    :keyword log_y: bool; whether to use a log scale on the y-axis (default is False)
    :keyword marginal_x: str: None, 'rug', 'box', 'violin', 'histogram'; type of marginal plot for the x-axis (default is "box")
    :keyword marginal_y: str: None, 'rug', 'box', 'violin', 'histogram'; type of marginal plot for the y-axis (default is "box")
    :keyword trendline: str: None, 'trace', 'overall'; type of trendline to add to the plot (default is None)
        if None - no trendline is added
        if 'trace' - a trendline is added for each trace
        if 'overall' - a single trendline is added for all traces
    :keyword trendline_color: str; color of the overall trendline (default is 'black'). has no effect if trendline is not 'overall'
    :keyword trendline_log_x: bool; whether to use a log scale on the x-axis of the trendline (default is False)
    :keyword trendline_log_y: bool; whether to use a log scale on the y-axis of the trendline (default is False)
    :keyword x_feature_units: str; units of the x-axis feature (default is "")
    :keyword y_feature_units: str; units of the y-axis feature (default is "")

    :return:
        - fig: go.Figure; the figure object
        - stat_results: pd.DataFrame; statistics of the trendline(s) (if any)
    """
    summary_df = summarize_events(events)
    if x_feature not in summary_df.columns:
        raise KeyError(f"Feature '{x_feature}' not found in the summary DataFrame.")
    if y_feature not in summary_df.columns:
        raise KeyError(f"Feature '{y_feature}' not found in the summary DataFrame.")
    if not kwargs.get("include_outliers", True):
        summary_df = summary_df[~summary_df["is_outlier"]]
    summary_df[cnst.LABEL_STR] = summary_df[cnst.LABEL_STR].map(
        lambda l: EventLabelEnum(l).name.replace("_", " ").title()
    )
    trendline = kwargs.get("trendline", None)
    if trendline is None:
        trendline_options = dict()
    elif trendline == "trace":
        trendline_options = dict(
            trendline="ols", trendline_scope="trace", trendline_options=dict(
                log_x=kwargs.get("trendline_log_x", False), log_y=kwargs.get("trendline_log_y", False),
            ))
    elif trendline == "overall":
        trendline_options = dict(
            trendline="ols", trendline_scope="overall", trendline_color_override=kwargs.get("trendline_color", 'black'),
            trendline_options=dict(
                log_x=kwargs.get("trendline_log_x", False), log_y=kwargs.get("trendline_log_y", False),
            ))
    else:
        raise ValueError(f"Invalid `trendline` argument: {trendline}")
    marg_x, marg_y = kwargs.get("marginal_x", "box"), kwargs.get("marginal_y", "box")
    fig = px.scatter(
        summary_df, x=x_feature, y=y_feature,
        color=cnst.LABEL_STR, symbol="is_outlier", symbol_map={True: "x", False: "circle"},
        # TODO: use color_discrete_map argument to set default event colors
        log_x=kwargs.get("log_x", False), marginal_x=marg_x, log_y=kwargs.get("log_y", False), marginal_y=marg_y,
        **trendline_options
    )
    stat_results = px.get_trendline_results(fig)

    # remove trendlines from marginal subplots (see https://stackoverflow.com/q/78746094/8543025):
    fig.update_traces(visible=False, selector=dict(name='Overall Trendline'))
    fig.update_traces(visible=True, showlegend=True, selector=dict(name='Overall Trendline', xaxis='x'))

    # update titles
    x_feature_title = x_feature.replace("_", " ").title()
    y_feature_title = y_feature.replace("_", " ").title()
    x_feature_units, y_feature_units = kwargs.get("x_feature_units", ""), kwargs.get("y_feature_units", "")
    fig.update_layout(
        title=title or f"{y_feature_title} vs. {x_feature_title}",
        xaxis_title=f"{x_feature_title} ({x_feature_units})" if x_feature_units else x_feature_title,
        yaxis_title=f"{y_feature_title} ({y_feature_units})" if y_feature_units else y_feature_title,
        legend=dict(xanchor='right', x=1),
    )
    return fig, stat_results



import os
import io
import itertools
import zipfile as zp
import posixpath as psx
from typing import final, List, Tuple, Dict, Union
from abc import ABC, abstractmethod


import numpy as np
import pandas as pd
import requests as req
from tqdm import tqdm
from scipy.io import loadmat
from scipy.interpolate import interp1d
import arff

import peyes._utils.constants as cnst
from peyes._utils.pixel_utils import calculate_pixel_size, visual_angle_to_pixels
from peyes._utils.event_utils import parse_label
from peyes._DataModels.EventLabelEnum import EventLabelEnum


class BaseDatasetLoader(ABC):
    _NAME: str
    _URL: str
    _ARTICLES: List[str]
    _LICENSE: str
    _INDEXERS: List[str] = [
        cnst.TRIAL_ID_STR, cnst.SUBJECT_ID_STR, cnst.STIMULUS_TYPE_STR, cnst.STIMULUS_NAME_STR
    ]

    @classmethod
    @final
    def load(cls, directory: str, save: bool = False, verbose: bool = False) -> pd.DataFrame:
        """
        Loads the dataset from the specified directory. If the dataset is not found, it is downloaded from the internet.
        If `save` is True and the dataset was downloaded, it is saved to the specified directory.
        if `verbose` is True, a progress bar is displayed while parsing the downloaded dataset.
        :return: a DataFrame containing the dataset
        :raises ValueError: if `save` is True and `directory` is not specified
        """
        if save and not directory:
            raise ValueError("Directory must be specified to save the dataset")
        try:
            dataset = pd.read_pickle(os.path.join(directory, f"{cls.name()}.pkl"))
        except (FileNotFoundError, TypeError) as _e:
            if verbose:
                if directory:
                    print(f"Dataset {cls.name()} not found in directory {directory}.")
                print("Downloading...")
            dataset = cls.download(verbose)
            if save:
                os.makedirs(directory, exist_ok=True)
                file_path = os.path.join(directory, f"{cls.name()}.pkl")
                if verbose:
                    print(f"Saved dataset to {file_path}")
                dataset.to_pickle(file_path)
        return dataset

    @classmethod
    @final
    def download(cls, verbose: bool = False) -> pd.DataFrame:
        """ Downloads the dataset from the internet, parses it and returns a DataFrame with cleaned data """
        url = cls.url()
        response = req.get(cls._URL)
        code = response.status_code
        if code != 200:
            raise ConnectionError(
                f"HTTP status code {code} when attempting to download dataset {cls.name()} from {url}"
            )
        df = cls._parse_response(response, verbose)
        return cls._reorder_columns(df)

    @classmethod
    @abstractmethod
    def _parse_response(cls, response: req.Response, verbose: bool = False) -> pd.DataFrame:
        """ Parses the downloaded response and returns a DataFrame containing the raw dataset """
        raise NotImplementedError

    @classmethod
    @final
    def _reorder_columns(cls, df: pd.DataFrame) -> pd.DataFrame:
        ordered_columns = sorted(df.columns, key=lambda col: cls.column_order().get(col, 10))
        return df[ordered_columns]

    @classmethod
    @final
    def name(cls) -> str:
        """ Name of the dataset """
        if not cls._NAME:
            raise AttributeError(f"Class {cls.__name__} must implement class attribute `_NAME`")
        return cls._NAME

    @classmethod
    @final
    def url(cls) -> str:
        """ URL of the dataset """
        if not cls._URL:
            raise AttributeError(f"Class {cls.__name__} must implement class attribute `_URL`")
        return cls._URL

    @classmethod
    @final
    def articles(cls) -> List[str]:
        """ List of articles to cite when using this dataset """
        if not cls._ARTICLES:
            raise AttributeError(f"Class {cls.__name__} must implement class attribute `_{cnst.ARTICLES_STR}`")
        return cls._ARTICLES

    @classmethod
    @final
    def license(cls) -> str:
        """ License of the dataset """
        if not cls._LICENSE:
            raise AttributeError(f"Class {cls.__name__} must implement class attribute `_LICENSE`")
        return cls._LICENSE

    @classmethod
    @final
    def documentation(cls) -> str:
        """ Returns a string with information about the dataset """
        title = f"Dataset:\t{cls.name().replace('_', ' ').title()}"
        lcns = f"License:\t{cls.license()}"
        urls = f"URL:\t{cls.url()}"
        articles = "Articles:\n" + "\n".join([f"- {a}" for a in cls.articles()])
        docstring = cls.__doc__ if cls.__doc__ else ""
        return f"{title}\n{lcns}\n{urls}\n{articles}\n\n{docstring}"

    @staticmethod
    def column_order() -> Dict[str, float]:
        return {
            cnst.TRIAL_ID_STR: 0.1, cnst.SUBJECT_ID_STR: 0.2, cnst.STIMULUS_TYPE_STR: 0.3, cnst.STIMULUS_NAME_STR: 0.4,
            cnst.T: 1.0, cnst.X: 1.1, cnst.Y: 1.2, cnst.PUPIL: 1.3,
            cnst.LEFT_X: 2.1, cnst.LEFT_Y: 2.2, cnst.LEFT_PUPIL: 2.3,
            cnst.RIGHT_X: 3.1, cnst.RIGHT_Y: 3.2, cnst.RIGHT_PUPIL: 3.3,
            cnst.PIXEL_SIZE_STR: 4.2, cnst.VIEWER_DISTANCE_STR: 4.3
        }

    @staticmethod
    @final
    def _extract_filename_and_extension(full_path: str) -> (str, str, str):
        """ Splits a full path into its components: path, filename and extension """
        path, extension = os.path.splitext(full_path)
        path, filename = os.path.split(path)
        return path, filename, extension


class Lund2013DatasetLoader(BaseDatasetLoader):
    """
    Loads the dataset from article: "One algorithm to rule them all? An evaluation and discussion of ten eye movement
    event-detection algorithms.", Andersson et al. (2017).

    This loader is based on a previous implementation, see article:
    Startsev, M., Zemblys, R. Evaluating Eye Movement Event Detection: A Review of the State of the Art. Behav Res 55, 1653–1714 (2023)
    See their implementation: https://github.com/r-zemblys/EM-event-detection-evaluation/blob/main/misc/data_parsers/lund2013.py
    
    ** Important Notes: **
    (1) The dataset used in Andersson et al. (2017) is a subset of the complete dataset, available through this loader.
    The subset of the dataset used in the article is available at: 'EyeMovementDetectorEvaluation-master/annotated_data/data used in the article/'
    (2) Two files had to be replaced due to errors in the original dataset. The corrected files are included in the
    dataset and used instead of the erroneous ones.
    """

    _NAME = "Lund2013"
    _URL = 'https://github.com/richardandersson/EyeMovementDetectorEvaluation/archive/refs/heads/master.zip'
    _LICENSE = "GNU GPL-3.0"
    _ARTICLES = [
        "Andersson, R., Larsson, L., Holmqvist, K., Stridh, M., & Nyström, M. (2017): One algorithm to rule them " +
        "all? An evaluation and discussion of ten eye movement event-detection algorithms. Behavior Research " +
        "Methods, 49(2), 616-637.",
    ]

    __PREFIX = 'EyeMovementDetectorEvaluation-master/annotated_data/originally uploaded data/'
    # note: the article contained only a subset of the full dataset. article data is in directory 'annotated_data/data used in the article/'
    __ERRONEOUS_FILES = ['UH29_img_Europe_labelled_MN.mat']
    __CORRECTION_FILES = [
        'EyeMovementDetectorEvaluation-master/annotated_data/fix_by_Zemblys2018/UH29_img_Europe_labelled_FIX_MN.mat'
    ]

    @classmethod
    def _parse_response(cls, response: req.Response, verbose: bool = False) -> pd.DataFrame:

        def is_valid_filename(name: str) -> bool:
            # check if the file is a valid mat file containing gaze data
            if not name.startswith(cls.__PREFIX):
                return False
            if not name.endswith('.mat'):
                return False
            if any(name.endswith(errs) for errs in cls.__ERRONEOUS_FILES):
                # skip erroneous files, see readme.md for more info
                return False
            return True

        # extract the zip file and list all files that contain gaze data
        zip_file = zp.ZipFile(io.BytesIO(response.content))
        file_names = [f for f in zip_file.namelist() if is_valid_filename(f)]
        file_names.extend(cls.__CORRECTION_FILES)

        # read all files into a list of dataframes
        trial_id = 0
        dataframes = {}
        for f in tqdm(file_names, desc="Processing Files", disable=not verbose):
            file = zip_file.open(f)
            gaze_data = cls.__read_eyetracker_data(file)
            subject_id, stimulus_type, stimulus_name, rater = cls.__extract_metadata(file)
            gaze_data.rename(columns={cnst.LABEL_STR: rater}, inplace=True)

            # write the DF to a dict based on the subject id, stimulus type, stimulus name, or add to existing DF
            existing_df = dataframes.get((subject_id, stimulus_type, stimulus_name), None)
            if existing_df is None:
                trial_id += 1
                gaze_data[cnst.TRIAL_ID_STR] = trial_id
                gaze_data[cnst.SUBJECT_ID_STR] = subject_id
                gaze_data[cnst.STIMULUS_TYPE_STR] = stimulus_type
                gaze_data[cnst.STIMULUS_NAME_STR] = stimulus_name
                dataframes[(subject_id, stimulus_type, stimulus_name)] = gaze_data
            else:
                existing_df.loc[:, rater] = gaze_data.loc[:, rater]
        return pd.concat(dataframes.values(), ignore_index=True, axis=0)

    @staticmethod
    def __read_eyetracker_data(file) -> pd.DataFrame:
        mat = loadmat(file)
        eyetracking_data = mat["ETdata"]
        eyetracking_data_dict = {name: eyetracking_data[name][0, 0] for name in eyetracking_data.dtype.names}

        # extract singleton values and convert from meters to cm:
        sampling_rate = eyetracking_data_dict['sampFreq'][0, 0]
        view_dist = eyetracking_data_dict['viewDist'][0, 0] * 100
        screen_width, screen_height = eyetracking_data_dict['screenDim'][0] * 100
        screen_res = eyetracking_data_dict['screenRes'][0]  # (1024, 768)
        pixel_size = calculate_pixel_size(screen_width, screen_height, screen_res)

        # extract gaze data:
        samples_data = eyetracking_data_dict['pos']
        right_x, right_y = samples_data[:, 3:5].T  # only recording right eye
        is_missing = (right_x == 0) & (right_y == 0)    # missing samples are marked with (0, 0) coordinates
        right_x[is_missing] = np.nan
        right_y[is_missing] = np.nan
        labels = pd.Series(samples_data[:, 5]).apply(lambda x: parse_label(x, safe=True))
        if np.isnan(samples_data[:, 0]).any():
            # if timestamps are NaN, re-populate them
            timestamps = np.arange(len(right_x)) * cnst.MILLISECONDS_PER_SECOND / sampling_rate
        else:
            # timestamps are available but in microseconds
            timestamps = samples_data[:, 0] - np.nanmin(samples_data[:, 0])  # start timestamps from 0
            timestamps /= cnst.MICROSECONDS_PER_MILLISECOND
        return pd.DataFrame(data={
            cnst.T: timestamps, cnst.X: right_x, cnst.Y: right_y, cnst.PUPIL: np.nan,
            cnst.LABEL_STR: labels, cnst.VIEWER_DISTANCE_STR: view_dist, cnst.PIXEL_SIZE_STR: pixel_size
        })

    @staticmethod
    def __extract_metadata(file) -> Tuple[str, str, Union[str, int], str]:
        file_name = os.path.basename(file.name)  # remove path
        if not file_name.endswith(".mat"):
            raise ValueError(f"Expected a `.mat` file, got: {file_name}")
        # file_name fmt: `<subject_id>_<stimulus_type>_<stimulus_name_1>_ ... _<stimulus_name_N>_labelled_<rater_name>`
        # moving-dot trials don't contain stimulus names
        file_name = file_name.replace(".mat", "")  # remove extension
        split_name = file_name.split("_")
        subject_id = split_name[0]  # subject id is always 1st in the file name
        rater = split_name[-1].upper()  # rater is always last in the file name
        stimulus_type = split_name[1]  # stimulus type is always 2nd in the file name
        if stimulus_type == cnst.VIDEO_STR:
            stimulus_name = "_".join(split_name[2:-2]).removesuffix("_labelled")  # between stim type and rater
            return subject_id, stimulus_type, stimulus_name, rater
        if stimulus_type == "img":
            stimulus_name = "_".join(split_name[2:-2]).removesuffix("_labelled")  # between stim type and rater
            stimulus_type = cnst.IMAGE_STR  # rename stimulus type to match peyes constants
            return subject_id, stimulus_type, stimulus_name, rater
        if stimulus_type.startswith(cnst.TRIAL_STR):
            # moving-dot stimulus is labelled as "trial1" or "trial17"
            stimulus_name = int(stimulus_type.removeprefix(cnst.TRIAL_STR))
            stimulus_type = cnst.MOVING_DOT_STR
            return subject_id, stimulus_type, stimulus_name, rater
        raise ValueError(f"Unknown stimulus type: {stimulus_type}")


class IRFDatasetLoader(BaseDatasetLoader):
    """
    Loads the dataset from a replication study of the article:
    Using machine learning to detect events in eye-tracking data. Zemblys et al. (2018).
    See also about the repro study: https://github.com/r-zemblys/irf/blob/master/doc/IRF_replication_report.pdf

    Note: binocular data was recorded but only one pair of (x, y) coordinates is provided.

    This loader is based on a previous implementation, see article:
    Startsev, M., Zemblys, R. Evaluating Eye Movement Event Detection: A Review of the State of the Art. Behav Res 55, 1653–1714 (2023)
    See their implementation: https://github.com/r-zemblys/EM-event-detection-evaluation/blob/main/misc/data_parsers/humanFixationClassification.py
    """

    _NAME = "IRF"
    _URL = r'https://github.com/r-zemblys/irf/archive/refs/heads/master.zip'
    _LICENSE = "MIT"
    _ARTICLES = [
        "Zemblys, Raimondas and Niehorster, Diederick C and Komogortsev, Oleg and Holmqvist, Kenneth. Using machine " +
        "learning to detect events in eye-tracking data. Behavior Research Methods, 50(1), 160–181 (2018)."
    ]

    __PREFIX = 'irf-master/etdata/lookAtPoint_EL'
    __STIMULUS_TYPE_VAL = "moving_dot"  # all subjects were shown the same 13-point moving dot stimulus
    __VIEWER_DISTANCE_CM_VAL = 56.5
    __MONITOR_WIDTH_CM_VAL, __MONITOR_HEIGHT_CM_VAL = 37.5, 30.2
    __MONITOR_RESOLUTION_VAL = (1280, 1024)
    __PIXEL_SIZE_CM_VAL = calculate_pixel_size(
        __MONITOR_WIDTH_CM_VAL, __MONITOR_HEIGHT_CM_VAL, __MONITOR_RESOLUTION_VAL
    )
    __RATER_NAME = "RZ"

    @classmethod
    def _parse_response(cls, response: req.Response, verbose: bool = False) -> pd.DataFrame:
        zip_file = zp.ZipFile(io.BytesIO(response.content))
        gaze_file_names = [
            f for f in zip_file.namelist() if
            (f.startswith(psx.join(cls.__PREFIX, "lookAtPoint_EL_")) and f.endswith('.npy'))
        ]
        gaze_dfs = []
        for i, f in tqdm(enumerate(gaze_file_names), desc="Processing Files", disable=not verbose):
            file = zip_file.open(f)
            _, file_name, _ = cls._extract_filename_and_extension(f)
            gaze_data = pd.DataFrame(np.load(file))
            gaze_data['evt'] = gaze_data['evt'].apply(lambda x: parse_label(x, safe=True))  # convert labels
            gaze_data[cnst.SUBJECT_ID_STR] = file_name.split('_')[-1]  # format: "lookAtPoint_EL_S<subject_num>"
            gaze_data[cnst.TRIAL_ID_STR] = i + 1
            gaze_data[cnst.PUPIL] = np.nan
            gaze_dfs.append(gaze_data)
        df = pd.concat(gaze_dfs, ignore_index=True, axis=0)

        # add metadata columns:
        df[cnst.STIMULUS_TYPE_STR] = cls.__STIMULUS_TYPE_VAL
        df[cnst.VIEWER_DISTANCE_STR] = cls.__VIEWER_DISTANCE_CM_VAL
        df[cnst.PIXEL_SIZE_STR] = cls.__PIXEL_SIZE_CM_VAL

        # remap columns to correct values
        df.rename(
            columns={"t": cnst.T, "evt": cls.__RATER_NAME, "x": cnst.X, "y": cnst.Y}, inplace=True
        )
        df[cnst.T] = df[cnst.T] * cnst.MILLISECONDS_PER_SECOND  # convert seconds to milliseconds
        df = cls.__correct_coordinates(df)
        return df

    @classmethod
    def __correct_coordinates(cls, df: pd.DataFrame) -> pd.DataFrame:
        new_df = df.copy()
        nan_idxs = new_df[~new_df[cnst.STATUS_STR]].index
        new_df.loc[nan_idxs, cnst.X] = np.nan
        new_df.loc[nan_idxs, cnst.Y] = np.nan
        new_df.drop(columns=[cnst.STATUS_STR], inplace=True)

        pixel_width_cm = cls.__MONITOR_WIDTH_CM_VAL / cls.__MONITOR_RESOLUTION_VAL[0]
        x = new_df[cnst.X].apply(
            lambda ang: visual_angle_to_pixels(
                angle=ang, d=cls.__VIEWER_DISTANCE_CM_VAL, pixel_size=pixel_width_cm, use_radians=False, keep_sign=True
            )
        )
        x += cls.__MONITOR_RESOLUTION_VAL[0] // 2  # move x=0 coordinate to the left of the screen
        pixel_height = cls.__MONITOR_HEIGHT_CM_VAL / cls.__MONITOR_RESOLUTION_VAL[1]
        y = new_df[cnst.Y].apply(
            lambda ang: visual_angle_to_pixels(
                angle=ang, d=cls.__VIEWER_DISTANCE_CM_VAL, pixel_size=pixel_height, use_radians=False, keep_sign=True
            )
        )
        y += cls.__MONITOR_RESOLUTION_VAL[1] // 2  # move y=0 coordinate to the top of the screen
        new_df.loc[:, cnst.X] = x.astype("float32")
        new_df.loc[:, cnst.Y] = y.astype("float32")
        return new_df


class HFCDatasetLoader(BaseDatasetLoader):
    """
    Loads the two datasets presented in articles:
    - (adults) Is human classification by experienced untrained observers a gold standard in fixation detection?
        Hooge et al. (2018)
    - (infants) An in-depth look at saccadic search in infancy.
        Hessels et al. (2016)

    This loader is based on a previous implementation, see article:
    Startsev, M., Zemblys, R. Evaluating Eye Movement Event Detection: A Review of the State of the Art. Behav Res 55, 1653–1714 (2023)
    See their implementation: https://github.com/r-zemblys/EM-event-detection-evaluation/blob/main/misc/data_parsers/humanFixationClassification.py

    Note the original publication did not specify the viewer distance used when recording the data. Here, we use the
    value from the Tobii TX300 eye-tracker's specifications: 65.0 cm (https://www.spectratech.gr/Web/Tobii/pdf/TX300.pdf).
    They used their eye-tracker's monitor with a resolution of 1920x1080 pixels, an aspect ratio of 16:9 and a diagonal
    of 23". Based on this site (https://max.pm/posts/screensize/) we calculate the height and width of the monitor to be
    28.64cm and 50.92cm, respectively. The pixel size is calculated based on these values.
    """

    _NAME: str = "HFC"
    _URL = r'https://github.com/dcnieho/humanFixationClassification/archive/refs/heads/master.zip'
    _LICENSE = "CC NC-BY-SA 4.0"
    _ARTICLES = [
        "Hooge, I.T.C., Niehorster, D.C., Nyström, M., Andersson, R. & Hessels, R.S. (2018). Is human classification " +
        "by experienced untrained observers a gold standard in fixation detection?",
    ]

    __PREFIX = 'humanFixationClassification-master/data'
    __SUBJECT_GROUP_STR = "subject_group"
    __INFANT_STR, __ADULT_STR = "infant", "adult"
    __SEARCH_TASK_STR, __FREE_VIEWING_STR = "search_task", "free_viewing"

    __VIEWER_DISTANCE_CM_VAL = 65
    __MONITOR_WIDTH_CM_VAL, __MONITOR_HEIGHT_CM_VAL = 50.92, 28.64
    __MONITOR_RESOLUTION_VAL = (1920, 1080)
    __PIXEL_SIZE_CM_VAL = calculate_pixel_size(
        __MONITOR_WIDTH_CM_VAL, __MONITOR_HEIGHT_CM_VAL, __MONITOR_RESOLUTION_VAL
    )

    @staticmethod
    def column_order() -> Dict[str, float]:
        return {
            **super(HFCDatasetLoader, HFCDatasetLoader).column_order(), HFCDatasetLoader.__SUBJECT_GROUP_STR: 6.1
        }

    @classmethod
    def _parse_response(cls, response: req.Response, verbose: bool = False) -> pd.DataFrame:
        zip_file = zp.ZipFile(io.BytesIO(response.content))
        # extract gaze data:
        gaze_file_names = [
            f for f in zip_file.namelist() if
            (f.startswith(psx.join(cls.__PREFIX, "ETdata")) and f.endswith('.txt'))
        ]
        gaze_dfs = {}
        for i, f in tqdm(enumerate(gaze_file_names), desc="Processing Files", disable=not verbose):
            file = zip_file.open(f)
            gaze_data = pd.read_csv(file, sep='\t')
            gaze_data[cnst.PUPIL] = np.nan  # no pupil data available
            _, file_name, _ = cls._extract_filename_and_extension(file.name)
            subject_group, subject_id = file_name.split('_')  # format: "<subject_type>_<subject_id>"
            gaze_data[cnst.SUBJECT_ID_STR] = subject_id
            gaze_data[cls.__SUBJECT_GROUP_STR] = subject_group
            gaze_data[cnst.STIMULUS_TYPE_STR] = cls.__FREE_VIEWING_STR if subject_group == cls.__ADULT_STR else cls.__SEARCH_TASK_STR
            gaze_data[cnst.TRIAL_ID_STR] = i + 1
            gaze_dfs[file_name] = gaze_data

        # extract annotations:
        coder_file_names = [
            f for f in zip_file.namelist() if
            (f.startswith(psx.join(cls.__PREFIX, "coderSettings")) and f.endswith('.txt'))
        ]
        annotation_dfs = {}
        for f in coder_file_names:
            with zip_file.open(f) as open_file:
                rater_data = pd.read_csv(open_file, sep='\t')
                _, rater_name, _ = cls._extract_filename_and_extension(open_file.name)
                annotation_dfs[rater_name.upper()] = rater_data

        # merge annotations with gaze data:
        merged_dfs = []
        for key, data in gaze_dfs.items():
            if data is None or len(data) == 0 or data.empty:
                continue
            l = len(data)
            for rater_name in annotation_dfs.keys():
                annotations = annotation_dfs.get(rater_name).query("Trial==@key")
                if annotations is None or len(annotations) == 0 or annotations.empty:
                    labels = np.zeros(l, dtype=int)
                else:
                    # reached here if there are annotations from this rater for this trial
                    labels = np.zeros(l, dtype=int)
                    for _, row in annotations.iterrows():
                        f = interp1d(
                            data["time"], range(l), kind="nearest", bounds_error=False, fill_value="extrapolate"
                        )
                        fixation_samples = itertools.chain(
                            *[range(int(s), int(e + 1)) for s, e in zip(
                                f(annotations["FixStart"]), f(annotations["FixEnd"])
                            )]
                        )
                        labels[list(fixation_samples)] = 1
                data[rater_name] = labels
                data[rater_name] = data[rater_name].apply(lambda x: parse_label(x, safe=True))
            merged_dfs.append(data)

        # concatenate all dataframes into a single one and add metadata columns
        full_dataset = pd.concat(merged_dfs, ignore_index=True, axis=0)
        full_dataset.rename(columns={"time": cnst.T, "x": cnst.X, "y": cnst.Y}, inplace=True)
        full_dataset[cnst.VIEWER_DISTANCE_STR] = cls.__VIEWER_DISTANCE_CM_VAL
        full_dataset[cnst.PIXEL_SIZE_STR] = cls.__PIXEL_SIZE_CM_VAL
        return full_dataset


class GazeComDatasetLoader(BaseDatasetLoader):
    """
    Loads a labelled subset of the dataset presented in the article:
    Michael Dorr, Thomas Martinetz, Karl Gegenfurtner, and Erhardt Barth. Variability of eye movements when viewing
    dynamic natural scenes. Journal of Vision, 10(10):1-17, 2010.

    Labels are from the article:
    Agtzidis, I., Startsev, M., & Dorr, M. (2016a). In the pursuit of (ground) truth: A hand-labelling tool for eye
    movements recorded during dynamic scene viewing. In 2016 IEEE second workshop on eye tracking and visualization
    (ETVIS) (pp. 65–68).

    Note 1: This dataset is extremely large and may take a long time to download. It is recommended to save it to local
    storage after downloading it for faster access in the future. The method `load_zipfile` can be used to load the
    dataset from a raw zip file stored in a local directory.
    Note 2: This is only a subset of the full GazeCom Dataset, containing hand-labelled samples. The full dataset with
    documentation can be found in https://www.inb.uni-luebeck.de/index.php?id=515.
    Note 3: binocular data was recorded but only one pair of (x, y) coordinates is provided.

    This loader is based on a previous implementation, see article:
    Startsev, M., Zemblys, R. Evaluating Eye Movement Event Detection: A Review of the State of the Art. Behav Res 55, 1653–1714 (2023)
    See their implementation: https://github.com/r-zemblys/EM-event-detection-evaluation/blob/main/misc/data_parsers/tum.py
    """

    _NAME: str = "GazeCom"
    _URL = r'https://gin.g-node.org/ioannis.agtzidis/gazecom_annotations/archive/master.zip'
    _LICENSE = "GNU GPL-3.0"
    _ARTICLES = [
        "Agtzidis, I., Startsev, M., & Dorr, M. (2016a). In the pursuit of (ground) truth: A hand-labelling tool for " +
        "eye movements recorded during dynamic scene viewing. In 2016 IEEE second workshop on eye tracking and" +
        "visualization (ETVIS) (pp. 65–68).",

        "Michael Dorr, Thomas Martinetz, Karl Gegenfurtner, and Erhardt Barth. Variability of eye movements when " +
        "viewing dynamic natural scenes. Journal of Vision, 10(10):1-17, 2010.",
        "Startsev, M., Agtzidis, I., & Dorr, M. (2016). Smooth pursuit. http://michaeldorr.de/smoothpursuit/",
    ]

    __PREFIX = psx.join('gazecom_annotations', 'ground_truth')
    __ZIPFILE_NAME = "gazecom_annotations-master.zip"
    __HANDLABELLER_PREFIX = "HL"
    __VIEWER_DISTANCE_CM_VAL = 56.5
    __MONITOR_WIDTH_CM_VAL, __MONITOR_HEIGHT_CM_VAL = 40, 22.5
    __MONITOR_RESOLUTION_VAL = (1280, 720)
    __PIXEL_SIZE_CM_VAL = calculate_pixel_size(
        __MONITOR_WIDTH_CM_VAL, __MONITOR_HEIGHT_CM_VAL, __MONITOR_RESOLUTION_VAL
    )

    __LABEL_MAP = {
        0: EventLabelEnum.UNDEFINED,
        1: EventLabelEnum.FIXATION,
        2: EventLabelEnum.SACCADE,
        3: EventLabelEnum.SMOOTH_PURSUIT,
        4: EventLabelEnum.UNDEFINED  # noise
    }
    __COLUMN_MAP = {
        "time": cnst.T, "x": cnst.X, "y": cnst.Y, "handlabeller1": f"{__HANDLABELLER_PREFIX}1",
        "handlabeller2": f"{__HANDLABELLER_PREFIX}2", "handlabeller_final": f"{__HANDLABELLER_PREFIX}_FINAL"
    }

    @classmethod
    def load_zipfile(cls, root: str = None, verbose: bool = False) -> pd.DataFrame:
        """
        Loads the dataset from a zip file stored in a local directory (so it doesn't need to be downloaded again).
        :param root: path to the directory containing the zip file.
        :param verbose: whether to display progress bars.
        :return: DataFrame with the annotated gaze data.
        """
        if not root or not psx.isdir(root):
            raise NotADirectoryError(f"Invalid directory: {root}")
        zip_file = psx.join(root, cls.__ZIPFILE_NAME)
        if not psx.isfile(zip_file):
            raise FileNotFoundError(f"File not found: {zip_file}")
        with zp.ZipFile(zip_file, 'r') as zip_ref:
            df = cls.__read_zipfile(zf=zip_ref, verbose=verbose)
            return cls._reorder_columns(df)

    @staticmethod
    def column_order() -> Dict[str, float]:
        handlabeller_scores = {
            f"{GazeComDatasetLoader.__HANDLABELLER_PREFIX}1": 5.1,
            f"{GazeComDatasetLoader.__HANDLABELLER_PREFIX}2": 5.2,
            f"{GazeComDatasetLoader.__HANDLABELLER_PREFIX}_FINAL": 5.3
        }
        return {**super(GazeComDatasetLoader, GazeComDatasetLoader).column_order(), **handlabeller_scores}

    @classmethod
    def _parse_response(cls, response: req.Response, verbose: bool = False) -> pd.DataFrame:
        zip_file = zp.ZipFile(io.BytesIO(response.content))
        return cls.__read_zipfile(zf=zip_file, verbose=verbose)

    @classmethod
    def __read_zipfile(cls, zf: zp.ZipFile, verbose: bool = False) -> pd.DataFrame:
        """
        Reads the contents of a zip file and returns a DataFrame with the annotated data.
        :param zf: ZipFile object
        :param verbose: whether to display progress bars
        :return: DataFrame with annotated gaze data
        """
        annotated_file_names = [f for f in zf.namelist() if (f.endswith('.arff') and cls.__PREFIX in f)]
        gaze_dfs = []
        for i, f in tqdm(enumerate(annotated_file_names), desc="Processing Files", disable=not verbose):
            file = zf.open(f)
            data = arff.loads(file.read().decode('utf-8'))

            # parse gaze data:
            df = pd.DataFrame(data['data'], columns=[attr[0] for attr in data['attributes']])
            df[cnst.PUPIL] = np.nan  # no pupil data available
            invalid_idxs = np.where(np.all(df[["x", "y"]] == 0, axis=1) | (df["confidence"] < 0.5))[0]
            df.iloc[invalid_idxs, df.columns.get_indexer(["x", "y"])] = np.nan
            df['time'] = df['time'] / cnst.MILLISECONDS_PER_SECOND
            df['time'] = df['time'] - df['time'].min()  # start timestamps from 0
            df.drop(columns=['confidence'], inplace=True)
            df.rename(columns=cls.__COLUMN_MAP, inplace=True)
            for col in df.columns:
                if col.startswith(cls.__HANDLABELLER_PREFIX):
                    df[col] = df[col].map(cls.__LABEL_MAP)

            # add metadata columns:
            _, file_name, _ = cls._extract_filename_and_extension(f)
            subj_id = file_name.split('_')[0]  # file_name: <subject_id>_<stimulus>_<name>_<with>_<underscores>.arff
            stimulus = '_'.join(file_name.split('_')[1:])
            df[cnst.SUBJECT_ID_STR] = subj_id
            df[cnst.STIMULUS_NAME_STR] = stimulus
            df[cnst.TRIAL_ID_STR] = i + 1
            gaze_dfs.append(df)

        # merge and add common metadata columns:
        full_dataset = pd.concat(gaze_dfs, ignore_index=True, axis=0)
        full_dataset[cnst.STIMULUS_TYPE_STR] = cnst.VIDEO_STR
        full_dataset[cnst.VIEWER_DISTANCE_STR] = cls.__VIEWER_DISTANCE_CM_VAL
        full_dataset[cnst.PIXEL_SIZE_STR] = cls.__PIXEL_SIZE_CM_VAL
        return full_dataset



import numpy as np
import pandas as pd

from peyes._utils.constants import MILLISECONDS_PER_SECOND
from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType
from peyes._DataModels.EventLabelEnum import EventLabelEnum

from peyes._utils.event_utils import parse_label as _parse_label
from peyes._utils.metric_utils import transition_matrix as _transition_matrix


def event_rate(
        events: EventSequenceType,
        label: UnparsedEventLabelType,
) -> float:
    """
    Calculates the number of occurrences of the given event-label per second.
    :param events: sequence of gaze-events
    :param label: event-label to calculate the rate for
    :return: the rate (Hz)
    """
    last_offset = events[-1].end_time   # last event's end-time in ms
    label = _parse_label(label)
    label_events = [e for e in events if e.label == label]
    events_to_ms = len(label_events) / last_offset
    return events_to_ms * MILLISECONDS_PER_SECOND


def microsaccade_rate(events: EventSequenceType, max_amplitude: float = 1.0) -> float:
    """
    Calculate the rate of micro-saccades per second.
    :param events: sequence of gaze-events
    :param max_amplitude: the maximum amplitude of a micro-saccade (deg)
    :return: the rate of micro-saccades per second (Hz)
    """
    assert max_amplitude > 0, "Micro-saccade threshold must be positive"
    microsaccades = [e for e in events if e.amplitude <= max_amplitude and e.label == EventLabelEnum.SACCADE]
    return len(microsaccades) / events[-1].end_time * MILLISECONDS_PER_SECOND


def microsaccade_ratio(events: EventSequenceType, max_amplitude: float = 1.0, zero_division: float = np.nan) -> float:
    """
    Calculate the ratio of micro-saccades to all saccades.
    Returns `zero_division` if there are no saccades.

    :param events: sequence of gaze-events
    :param max_amplitude: the maximum amplitude of a micro-saccade (deg)
    :param zero_division: value to return if there are no saccades
    :return: the ratio of micro-saccades to all saccades
    """
    assert max_amplitude > 0, "Micro-saccade threshold must be positive"
    saccades = [e for e in events if e.label == EventLabelEnum.SACCADE]
    microsaccades = [e for e in saccades if e.amplitude <= max_amplitude]
    try:
        return len(microsaccades) / len(saccades)
    except ZeroDivisionError:
        return zero_division


def transition_matrix(
        seq: EventSequenceType,
        normalize_rows: bool = False
) -> pd.DataFrame:
    """
    Calculates the transition matrix from a sequence of event.
    If `normalize_rows` is True, the matrix will be normalized by the sum of each row, i.e. contains transition probabilities.
    Returns a DataFrame where rows indicate the origin event-label and columns indicate the destination event-label.
    """
    return _transition_matrix([e.label for e in seq], normalize_rows)
